package fi.helsinki.cs.seco.onki.index;

import java.io.IOException;
import java.text.Normalizer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Collection;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.io.File;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.ConstantScoreRangeQuery;
import org.apache.lucene.search.Hit;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopFieldDocs;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.WildcardQuery;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.search.WildcardTermEnum;
import org.apache.lucene.search.TermsFilter;
import org.apache.lucene.index.IndexReader;

/**
 * A Lucene-based alphabetical index for concepts. The index is kept either in RAM or
 * in a directory in the file system.
 * @author oisuomin
 */

public class AlphabeticalIndex implements IAlphabeticalIndex {
	
	public static final String GROUP = "group";
	public static final String ROOT = "root";
	
	private Directory dir;
	private boolean newIndex;
	private Set<String> knownlangs;
	private Set<String> knowntypes;
	// should there be a separate map of knownPrefLabelAlphabets (for REST API's getDirectory)?
	private Map<String, Set<Character>> knownalphabets;
	private Map<String, Map<String, Integer>> knownLabelPrefixes;
	private Set<String> knownNamespaces;
	private String path;
	private IndexSearcher is;
	private String defaultSortingLang;
	
	/**
	 * Create a RAM-based non-persistent index.
	 */
	
	public AlphabeticalIndex(String defaultSortingLang) {
		this.dir = new RAMDirectory();
		this.newIndex = true;
		this.knownlangs = new HashSet<String>();
		this.knowntypes = new HashSet<String>();
		this.knownalphabets = new HashMap<String, Set<Character>>();
		this.knownLabelPrefixes = new HashMap<String, Map<String, Integer>>();
		this.knownNamespaces = new HashSet<String>();
		this.defaultSortingLang = defaultSortingLang;
	}

	/**
	 * Create a directory-based persistent index. The directory will be created if it doesn't exist.
	 */

	public AlphabeticalIndex(String path,String defaultSortingLang) {
		this(path, -1, defaultSortingLang);
	}
	
	/**
	 * Create a directory-based persistent index. The directory will be created if it doesn't exist
	 * or if the ontology has been modified since the time of creation of the index directory.
	 */
	
	public AlphabeticalIndex(String path, long lastModified, String defaultSortingLang) {
		this.path = path;
		this.newIndex = true;
		try {
			File dirFile = new File(path);
			if (dirFile.exists() && dirFile.isDirectory() && lastModified != -1 && dirFile.lastModified() > lastModified)
				this.newIndex = false;
			
			this.dir = FSDirectory.getDirectory(path);
			
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		this.knownlangs = new HashSet<String>();
		this.knowntypes = new HashSet<String>();
		this.knownalphabets = new HashMap<String, Set<Character>>();
		this.knownLabelPrefixes = new HashMap<String, Map<String, Integer>>();
		this.knownNamespaces = new HashSet<String>();
		this.defaultSortingLang = defaultSortingLang;
	}

	/**
	 * Initialize the index with a set of items. The directory contents, if any, will be overwritten
	 * if the ontology has been modified since the creation of the index.
	 */
	
	public void initialize(Iterable<IndexItem> items) {
		this.initialize(items, null);
	}
	
    /**
     * Initialize the index with a set of items. The directory contents, if any, will be overwritten
	 * if the ontology has been modified since the creation of the index. Only resources from the defined
	 * are indexed.
	 * If the index is newer than the ontology, the existing index is used and the languages
	 * of concept labels and types of concepts are fetched from the index.  
	 * 
	 */
	
	public void initialize(Iterable<IndexItem> items, Set<String> allowedNamespaces) {
		try {
			if (this.newIndex) {
				try {
					IndexWriter iw = new IndexWriter(this.dir, new StandardAnalyzer(), true);

					/*if (path != null) {
						iw.setMergeFactor(1000);
						iw.setMaxMergeDocs(10000);
					}*/

					this.knownalphabets.put(null, new TreeSet<Character>());
					this.knownLabelPrefixes.put(null, new TreeMap<String, Integer>());
					
					int counter = 0;
					for (IndexItem item : items) {
						if (allowedNamespaces != null && allowedNamespaces.size() > 0
							&& !allowedNamespaces.contains(item.getUriNamespace())) {
							boolean filterOut = true;
							for (String ns : allowedNamespaces) {
								if (ns.endsWith("*") &&	item.getUriNamespace().startsWith(ns.substring(0, ns.length()-1))) {
									filterOut = false;
									break;
								}
							}
							if (filterOut)
								continue;
						}

						if (counter % 1000 == 0)
							System.out.println(item); // print out a sample of indexitems for testing purposes
						
						// index using prefLabel
						HashMap<String,String> docvalues = new HashMap<String,String>();
						docvalues.put("label", item.getPrefLabel().trim());
						docvalues.put("prefLabel", item.getPrefLabel().trim());
						docvalues.put("origLabel", item.getPrefLabel().trim());
						iw.addDocument(createDoc(item, docvalues));
						
						// index also the original, unnormalized label, so the queries containing diacritics
						// return only hits with labels containing diacritics
						String normalizedLabel = this.normalizeUnicodeString(item.getPrefLabel().trim());
						if (!normalizedLabel.equals(item.getPrefLabel().trim().toLowerCase())) {
							docvalues.put("label", normalizedLabel);
							iw.addDocument(createDoc(item, docvalues));
						}

						if (!this.knownalphabets.containsKey(item.getLanguage()))
							this.knownalphabets.put(item.getLanguage(), new TreeSet<Character>());
						
						if (normalizedLabel.length() > 0) {
							this.knownalphabets.get(item.getLanguage()).add(normalizedLabel.charAt(0));
							this.knownalphabets.get(null).add(normalizedLabel.charAt(0));
						}
						
						if (!this.knownLabelPrefixes.containsKey(item.getLanguage()))
							this.knownLabelPrefixes.put(item.getLanguage(), new TreeMap<String, Integer>());
						
						addLabelIntoKnownLabelPrefixes(item.getPrefLabel(), item.getLanguage());
						
						// index using each hiddenLabel
						for (String label : item.getHiddenLabels()) {
							docvalues.put("label", label.trim());
							iw.addDocument(createDoc(item, docvalues));
							
							// index also the original, unnormalized label, so the queries containing diacritics
							// return only hits with labels containing diacritics
							normalizedLabel = this.normalizeUnicodeString(label.trim());
							if (!normalizedLabel.equals(label.trim().toLowerCase())) {
								docvalues.put("label", normalizedLabel);
								iw.addDocument(createDoc(item, docvalues));
							}
							
							if (normalizedLabel.length() > 0) {
								this.knownalphabets.get(item.getLanguage()).add(normalizedLabel.charAt(0));
								this.knownalphabets.get(null).add(normalizedLabel.charAt(0));
							}
							
							addLabelIntoKnownLabelPrefixes(label, item.getLanguage());
						}
						
						// index using each altLabel
						for (String label : item.getAltLabels()) {
							docvalues.put("label", label.trim());
							docvalues.put("origLabel", label.trim());
							iw.addDocument(createDoc(item, docvalues));
							
							// index also the original, unnormalized label, so the queries containing diacritics
							// return only hits with labels containing diacritics
							normalizedLabel = this.normalizeUnicodeString(label.trim());
							if (!normalizedLabel.equals(label.trim().toLowerCase())) {
								docvalues.put("label", normalizedLabel);
								iw.addDocument(createDoc(item, docvalues));
							}
							
							if (normalizedLabel.length() > 0) {
								this.knownalphabets.get(item.getLanguage()).add(normalizedLabel.charAt(0));
								this.knownalphabets.get(null).add(normalizedLabel.charAt(0));
							}
							
							addLabelIntoKnownLabelPrefixes(label, item.getLanguage());
						}
						
						this.knownlangs.add(item.getLanguage());
						this.knownNamespaces.add(item.getUriNamespace());
						counter++;
					}

					iw.optimize();
					iw.close();
				} catch (IOException e) {
					// if indexing causes IOException the directory-based index is deleted, so the index is not left in unstable state
					if (dir instanceof FSDirectory) {
						File dirFile = new File(this.path);
						deleteDir(dirFile);
					}
					throw new RuntimeException(e);
				}
			}
			// if index already existed (on disk), knowlangs, knowntypes, etc. have to be fetched from index
			else {
				IndexSearcher is = new IndexSearcher(this.dir);
				IndexReader ir = is.getIndexReader();

				TermEnum langEnum = ir.terms(new Term("language", ""));
				do {
					if (langEnum.term() == null || langEnum.term().toString().indexOf("language") != 0)
						continue;
					this.knownlangs.add(langEnum.term().toString().substring(9));
				}
				while (langEnum.next());
				langEnum.close();				
				
				TermEnum typeEnum = ir.terms(new Term("type", ""));
				do {
					if (typeEnum.term() == null || typeEnum.term().toString().indexOf("type") != 0)
						continue;
					if (!typeEnum.term().toString().substring(5).equals(GROUP))
						this.knowntypes.add(typeEnum.term().toString().substring(5));
				}
				while (typeEnum.next());
				typeEnum.close();
					
				// temporal solution with no lang codes, Jouni 25.6.2009
				this.knownalphabets.put(null, new TreeSet<Character>());
				
				this.knownLabelPrefixes.put(null, new TreeMap<String, Integer>());
				
				TermEnum alphabetEnum = ir.terms(new Term("label", ""));
				do {
					if (alphabetEnum.term() == null || alphabetEnum.term().toString().indexOf("label") != 0 || alphabetEnum.term().toString().length() < 7)
						continue;
					String normalizedLabel = this.normalizeUnicodeString(alphabetEnum.term().toString().substring(6));
					if (normalizedLabel.length() > 0)
						this.knownalphabets.get(null).add(normalizedLabel.charAt(0));
					
					String label = alphabetEnum.term().toString().substring(6);
					if (label.length() > 2)
						label = label.substring(0, 3);
					int amount = 1;
					if (this.knownLabelPrefixes.get(null).containsKey(label))
						amount = this.knownLabelPrefixes.get(null).get(label)+1;
					this.knownLabelPrefixes.get(null).put(label, amount);
				}
				while (alphabetEnum.next());
				alphabetEnum.close();
				
				TermEnum uriEnum = ir.terms(new Term("uri", ""));
				do {
					if (uriEnum.term() == null || uriEnum.term().toString().indexOf("uri") != 0)
						continue;
					String uriNs = uriEnum.term().toString().substring(4);
					if (uriNs.indexOf('#') != -1)
						uriNs = uriNs.substring(0, uriNs.lastIndexOf('#')+1);
					else
						uriNs = uriNs.substring(0, uriNs.lastIndexOf('/')+1);
					this.knownNamespaces.add(uriNs);
				}
				while (uriEnum.next());
				uriEnum.close();
				
			}
			this.is = new IndexSearcher(this.dir);
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
	
    /**
	 * Remove diacritics from concept labels. 
	 */
    
	private String normalizeUnicodeString(String origString) {
		origString = origString.toLowerCase();
		char[] origStringArray = origString.toCharArray();
		char[] normalizedStringArray = Normalizer.normalize(origString, Normalizer.Form.NFD).replaceAll("[^\\p{ASCII}]", "").toCharArray();

		// TODO: move this exclusion list to conf
		LinkedList<Character> notNormalizedChars = new LinkedList<Character>();
		notNormalizedChars.add('å');
		notNormalizedChars.add('ä');
		notNormalizedChars.add('ö');
		
		// FIXME: in case the original and normalized label lengths do not match, characters that are not supposed
		// to be normalized (e.g. å, ä and ö) are normalized, because they cannot be preserved with the current processing
		if (origStringArray.length == normalizedStringArray.length) {
			for (int i=0; i<origStringArray.length; i++) {
				if (notNormalizedChars.contains(origStringArray[i]))
					normalizedStringArray[i] = origStringArray[i];
			}
		}
		// debug
		/*else {
			System.out.println("Warning: Label length mismatch while normalizing labels for search index: "+origString+" - "+new String(normalizedStringArray));
			for (char c : origStringArray)
				System.out.print(c+", ");
			System.out.println("");
			System.out.println("-");
			for (char c : normalizedStringArray)
				System.out.print(c+", ");
			System.out.println("");
		}*/
		
		return new String(normalizedStringArray).replaceAll("[\\s]+[-,.]*[\\s]*", " ").replaceAll("[\\s]*[-,.]*[\\s]+", " ");
	}
	
    /**
	 * Create index document from concept information.
	 */
	
	private Document createDoc(IndexItem item, Map<String,String> docvalues) {
		Document doc = new Document();
		
		doc.add(new Field("uri", item.getUri(), Field.Store.YES, Field.Index.UN_TOKENIZED));
		doc.add(new Field("uriNamespace", item.getUriNamespace(), Field.Store.YES, Field.Index.NO));
		doc.add(new Field("uriNamespacePrefix", item.getUriNamespacePrefix(), Field.Store.YES, Field.Index.NO));
		doc.add(new Field("uriLocalname", item.getUriLocalname(), Field.Store.YES, Field.Index.NO));
		doc.add(new Field("language", item.getLanguage(), Field.Store.YES, Field.Index.UN_TOKENIZED));
		
		//ISOLatin1AccentFilter filter = new ISOLatin1AccentFilter(new KeywordTokenizer(new StringReader(docvalues.get("label"))));
		//try {
			//System.out.println(filter.next().termText());
			//final Token reusableToken = new Token();
			//System.out.println(new String(filter.next(reusableToken).termBuffer()));
			//char[] test = "šamaanit".toCharArray();
			//filter.removeAccents(test, test.length);
			//System.out.println(test);
			//doc.add(new Field("label", filter.next().toString(), Field.Store.NO, Field.Index.UN_TOKENIZED));
		//}
		//catch (Exception e) {
			//System.out.println(e+" when removing diacritics.");;
		//}
		
		//System.out.println("NFC: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFC)
		//		+ "; NON-ASCII: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFC).replaceAll("[^\\p{ASCII}]", ""));
		//System.out.println("NFD: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFD)
		//		+ "; NON-ASCII: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFD).replaceAll("[^\\p{ASCII}]", ""));
		//System.out.println("NFKC: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFKC)
		//		  	+ "; NON-ASCII: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFKC).replaceAll("[^\\p{ASCII}]", ""));
		//System.out.println("NFKD: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFKD)
		//		  	+ "; NON-ASCII: "+Normalizer.normalize(docvalues.get("label").toLowerCase(), Normalizer.Form.NFKD).replaceAll("[^\\p{ASCII}]", ""));
				
		doc.add(new Field("label", docvalues.get("label").toLowerCase(), Field.Store.NO, Field.Index.UN_TOKENIZED));
		doc.add(new Field("prefLabel", docvalues.get("prefLabel"), Field.Store.YES, Field.Index.NO));
		doc.add(new Field("origLabel", docvalues.get("origLabel"), Field.Store.YES, Field.Index.NO));
		
		for (String type : item.getTypes()) { 
			doc.add(new Field("type", type, Field.Store.NO, Field.Index.UN_TOKENIZED));
			this.knowntypes.add(type);
		}
		
		if (item.isGroup())
			doc.add(new Field("type", GROUP, Field.Store.NO, Field.Index.UN_TOKENIZED));
		
		for (String superclass : item.getSuperClasses())
			doc.add(new Field("superclass", superclass, Field.Store.NO, Field.Index.UN_TOKENIZED));

		if (doc.getFields("superclass") == null) {		
			doc.add(new Field("superclass", ROOT, Field.Store.NO, Field.Index.UN_TOKENIZED));
		}
			
		for (String group : item.getGroups())
			doc.add(new Field("group", group, Field.Store.NO, Field.Index.UN_TOKENIZED));
		
		for (String queryExpansion : item.getQueryExpansions())
			doc.add(new Field("expandQuery", queryExpansion, Field.Store.NO, Field.Index.UN_TOKENIZED));
		
		return doc;
	}

    /**
	 * Get the upper bound for a range query (for autocompletion and matching the language tags).
	 */
	
	private String nextPrefix(String prefix) {
		int len = prefix.length();
		if (len == 0)
			return null;
		return prefix.substring(0, len-1) + new String(Character.toChars(prefix.charAt(len - 1) + 1));
	}

    /**
	 * Methods for searching the index.
	 */
	
	public IndexResults search(String language, String prefix) {
		return this.search(language, prefix, null, null, null, false, false, -1);
	}

	public IndexResults search(String language, String prefix, Collection<String> types, Collection<String> parents, Collection<String> groups, boolean isGroup, boolean isRoot, int maxHits) {
		return this.search(language, prefix, types, parents, groups, isGroup, isRoot, maxHits, -1);
	}
	
	public IndexResults search(String language, String prefix, Collection<String> types, Collection<String> parents, Collection<String> groups, boolean isGroup, boolean isRoot, int maxHits, int startFromHit) {
		return this.search(language, prefix, types, parents, groups, isGroup, isRoot, maxHits, startFromHit, null, null, null, false);
	}
	
	public IndexResults search(String language, String prefix, Collection<String> types, Collection<String> parents, Collection<String> groups, boolean isGroup, boolean isRoot, int maxHits, int startFromHit, Collection<String> excludedTypes, Collection<String> excludedParents, Collection<String> excludedGroups, boolean multipleHitsOfConcept) {
		// the query term is not normalized so that queries containing diacritics return only hits to labels
		// that contain those diacritics
		//prefix = this.normalizeUnicodeString(prefix);
		
		// don't try to search empty index
		if (this.is.getIndexReader().numDocs() == 0)
			return new IndexResults(new IndexResultsMetadata(0,0), new ArrayList<IndexResult>(0));
		
		try {
			//float time = System.currentTimeMillis();
			BooleanQuery bq = new BooleanQuery();
			TermsFilter filter = null;

			if (language != null) {
				if (!language.equals("")) {
					BooleanQuery langQuery = new BooleanQuery();
					langQuery.add(new TermQuery(new Term("language", language)), BooleanClause.Occur.SHOULD);
					langQuery.add(new ConstantScoreRangeQuery("language", language+"-", nextPrefix(language+"-"), true, false), BooleanClause.Occur.SHOULD);
					langQuery.add(new TermQuery(new Term("language", "")), BooleanClause.Occur.SHOULD);
					bq.add(langQuery, BooleanClause.Occur.MUST);
				}
				else
					bq.add(new TermQuery(new Term("language", "")), BooleanClause.Occur.MUST);
			}
			else // to prevent queries containing only MUST_NOT clause(s) (Lucene does not support them). quite slow.
				bq.add(new WildcardQuery(new Term("language", "*")), BooleanClause.Occur.MUST);
			
			if (types != null && types.size() > 0) {
				BooleanQuery typeQuery = new BooleanQuery();
				for (String type : types)
					typeQuery.add(new TermQuery(new Term("type", type)), BooleanClause.Occur.SHOULD);
				bq.add(typeQuery, BooleanClause.Occur.MUST);
			}
			if (excludedTypes != null && excludedTypes.size() > 0) {
				for (String excludedType : excludedTypes)
					bq.add(new TermQuery(new Term("type", excludedType)), BooleanClause.Occur.MUST_NOT);
			}
			
			if (parents != null && parents.size() > 0) {
				BooleanQuery parentQuery = new BooleanQuery();
				for (String parent : parents)
					parentQuery.add(new TermQuery(new Term("superclass", parent)), BooleanClause.Occur.SHOULD);
				bq.add(parentQuery, BooleanClause.Occur.MUST);
			}
			if (excludedParents != null && excludedParents.size() > 0) {
				for (String excludedParent : excludedParents)
					bq.add(new TermQuery(new Term("superclass", excludedParent)), BooleanClause.Occur.MUST_NOT);
			}
			
			if (groups != null && groups.size() > 0) {
				BooleanQuery groupQuery = new BooleanQuery();
				for (String group : groups)
					groupQuery.add(new TermQuery(new Term("group", group)), BooleanClause.Occur.SHOULD);
				bq.add(groupQuery, BooleanClause.Occur.MUST);
			}
			if (excludedGroups != null && excludedGroups.size() > 0) {
				for (String excludedGroup : excludedGroups)
					bq.add(new TermQuery(new Term("group", excludedGroup)), BooleanClause.Occur.MUST_NOT);
			}
			
			if (isGroup)
				bq.add(new TermQuery(new Term("type", GROUP)), BooleanClause.Occur.MUST);
			else
				bq.add(new TermQuery(new Term("type", GROUP)), BooleanClause.Occur.MUST_NOT);
			
			if (isRoot)
				bq.add(new TermQuery(new Term("superclass", ROOT)), BooleanClause.Occur.MUST);
			
			int length = prefix.length();
			
			// wildcard query
			if ((prefix.indexOf('?') > -1 && (prefix.indexOf('?') < length-1 || prefix.length()==1)) ||
				(prefix.indexOf('*') > -1 && (prefix.indexOf('*') < length-1 || prefix.length()==1))) {
				if (prefix.charAt(length-1) != '?' && prefix.charAt(length-1) != '*')
					prefix += "*";
				else if (prefix.charAt(length-1) == '?')
					prefix = prefix.substring(0, length-1) + "*";
				WildcardTermEnum termEnum = new WildcardTermEnum(is.getIndexReader(), new Term("label", prefix.toLowerCase()));
				filter = new TermsFilter();
				do {
					filter.addTerm(termEnum.term());
				}
				while (termEnum.next()); 
				termEnum.close();
			}
			// regular query
			else {
				if (length > 0 && (prefix.charAt(length-1) == '*' || prefix.charAt(length-1) == '?'))
					prefix = prefix.substring(0, length-1);
				if (prefix.equals("#")) {
					bq.add(new ConstantScoreRangeQuery("label", "0", "a", true, false), BooleanClause.Occur.MUST);
				}
				else
					bq.add(new ConstantScoreRangeQuery("label", prefix.toLowerCase(), nextPrefix(prefix.toLowerCase()), true, false), BooleanClause.Occur.MUST);
			}

			if (maxHits < 0)
				maxHits = 1000000;
			int queryMaxHits = maxHits;
			if (startFromHit > -1)
				queryMaxHits = 1000000;
			if (startFromHit == 0)
				startFromHit = 1;

			TopFieldDocs hits = this.is.search(bq, filter, queryMaxHits, new Sort("label"));

			HashMap<String, IndexResult> uriResultMap = new HashMap<String, IndexResult>();
			List<IndexResult> l = new LinkedList<IndexResult>();
			for (ScoreDoc scoreDoc : hits.scoreDocs) {
				Document doc = this.is.doc(scoreDoc.doc);
				// these bugs/features are present only when not using paging:
				// FIXME: if some hits are ignored, the resulting result set is smaller than maxHits
				// TODO: also the totalHits amount is not right value
				//       (if the maxHits is smaller that the amount of matching concepts in index)
				if (multipleHitsOfConcept || !uriResultMap.containsKey(doc.get("uri")) || (uriResultMap.get(doc.get("uri")).getAltLabel() != null && doc.get("prefLabel").equals(doc.get("origLabel")))) {
					String uriNamespace = doc.get("uriNamespace");
					String uriNamespacePrefix = doc.get("uriNamespacePrefix");
					String uriLocalname = doc.get("uriLocalname");
					String prefLabel = doc.get("prefLabel");
					String origLabel = doc.get("origLabel");
					String altLabel = (prefLabel.equals(origLabel)) ? null : origLabel;

					IndexResult res = new IndexResult(uriNamespace, uriNamespacePrefix, uriLocalname, prefLabel, altLabel);
					if (multipleHitsOfConcept)
						l.add(res);
					else					
						uriResultMap.put(doc.get("uri"), res);
				}
			}
			if (!multipleHitsOfConcept)
				l = new ArrayList<IndexResult>(uriResultMap.values());

			//System.out.println("AlphabeticalIndex.search(\"" + prefix + "\"@" + language + "): " + l.size() + " results, query took " + (System.currentTimeMillis() - time) + " ms");
			int totalHits = hits.totalHits;

			// if paging is used, correct the totalHits value
			if (startFromHit > -1)
				totalHits = l.size();
			
			// sort the hits using the ICU collator
			if (language == null)
				language = this.defaultSortingLang;
			IndexResultComparator comparator = new IndexResultComparator(language);
			Collections.sort(l, comparator);
			
			if (!l.isEmpty() && startFromHit > -1) {
				if (startFromHit > l.size())
					l = new ArrayList<IndexResult>();
				else {
					int endIndex = startFromHit-1+maxHits;
					if (endIndex > l.size())
						endIndex = l.size();
					l = l.subList(startFromHit-1, endIndex);
				}
			}

			int containingHits = l.size();
			
			// if paging is not used and all the results fit in the maxHits,
			// correct the totalHits value
			if (totalHits==hits.scoreDocs.length && startFromHit == -1)
				totalHits = containingHits;
			
			IndexResultsMetadata metadata = new IndexResultsMetadata(totalHits, containingHits);
			return new IndexResults(metadata, l);
		} catch (NullPointerException e) { // lucene throws NullPointerException, if wildcard query returns 0 hits, Jouni 24.1.2008
			return new IndexResults(new IndexResultsMetadata(0,0), new ArrayList<IndexResult>(0));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
	
    /**
	 * Perform query expansion for a concept.
	 */
	
	public IndexResults expandQuery(Collection<String> uris, String language, int maxHits, Collection<String> types) {
		try {
			BooleanQuery bq = new BooleanQuery();
			TermsFilter filter = null;
				
			BooleanQuery langQuery = new BooleanQuery();
			BooleanQuery typeQuery = new BooleanQuery();
			
			if (language != null) {
				if (!language.equals("")) {
					langQuery.add(new TermQuery(new Term("language", language)), BooleanClause.Occur.SHOULD);
					langQuery.add(new ConstantScoreRangeQuery("language", language+"-", nextPrefix(language+"-"), true, false), BooleanClause.Occur.SHOULD);
					langQuery.add(new TermQuery(new Term("language", "")), BooleanClause.Occur.SHOULD);
					bq.add(langQuery, BooleanClause.Occur.MUST);
				}
				else
					bq.add(new TermQuery(new Term("language", "")), BooleanClause.Occur.MUST);
			}

			if (types != null && types.size() > 0) {
				for (String type : types)
					typeQuery.add(new TermQuery(new Term("type", type)), BooleanClause.Occur.SHOULD);
				bq.add(typeQuery, BooleanClause.Occur.MUST);
			}
			
			//search uris
			if (uris != null && uris.size() > 0) {
				BooleanQuery uriQuery = new BooleanQuery();
				for (String uri : uris) {
					uriQuery.add(new TermQuery(new Term("uri", uri)), BooleanClause.Occur.SHOULD);
				}
				bq.add(uriQuery, BooleanClause.Occur.MUST);
			}		
						
			if (maxHits < 0)
				maxHits = 1000000;
			
			TopFieldDocs hits = this.is.search(bq, filter, maxHits, new Sort("label"));

			List<IndexResult> l = new ArrayList<IndexResult>();

			for (ScoreDoc scoreDoc : hits.scoreDocs) {
				Document doc = this.is.doc(scoreDoc.doc);
				//TODO: make a helper method for this IndexResult generation (used at least in 3 contexts)
				String uriNamespace = doc.get("uriNamespace");
				String uriNamespacePrefix = doc.get("uriNamespacePrefix");
				String uriLocalname = doc.get("uriLocalname");
				String prefLabel = doc.get("prefLabel");
				String origLabel = doc.get("origLabel");
				String altLabel = (prefLabel.equals(origLabel)) ? null : origLabel;

				IndexResult res = new IndexResult(uriNamespace, uriNamespacePrefix, uriLocalname, prefLabel, altLabel);
				l.add(res);
			}
			
			// search expanded uris
			maxHits = maxHits - hits.totalHits;
			if (maxHits > 0) {
				bq = new BooleanQuery();
				if (language != null)
					bq.add(langQuery, BooleanClause.Occur.MUST);
				else
					bq.add(new TermQuery(new Term("language", "")), BooleanClause.Occur.MUST);
				if (types != null && types.size() > 0)
					bq.add(typeQuery, BooleanClause.Occur.MUST);
				
				if (uris != null && uris.size() > 0) {
					BooleanQuery expandQuery = new BooleanQuery();
					for (String uri : uris) {
						expandQuery.add(new TermQuery(new Term("expandQuery", uri)), BooleanClause.Occur.SHOULD);
					}
					bq.add(expandQuery, BooleanClause.Occur.MUST);
				}			
	
				hits = this.is.search(bq, filter, maxHits, new Sort("label"));
	
				for (ScoreDoc scoreDoc : hits.scoreDocs) {
					Document doc = this.is.doc(scoreDoc.doc);
					// don't add the uri itself, it's already added in the previous step
					if (!uris.contains(doc.get("uri"))) {
						String uriNamespace = doc.get("uriNamespace");
						String uriNamespacePrefix = doc.get("uriNamespacePrefix");
						String uriLocalname = doc.get("uriLocalname");
						String prefLabel = doc.get("prefLabel");
						String origLabel = doc.get("origLabel");
						String altLabel = (prefLabel.equals(origLabel)) ? null : origLabel;
		
						IndexResult res = new IndexResult(uriNamespace, uriNamespacePrefix, uriLocalname, prefLabel, altLabel);
						l.add(res);
					}
				}
			}

			IndexResultsMetadata metadata = new IndexResultsMetadata(hits.totalHits, hits.scoreDocs.length);
			return new IndexResults(metadata, l);
		} catch (NullPointerException e) { // lucene throws NullPointerException, if wildcard query returns 0 hits, Jouni 24.1.2008
			return new IndexResults(new IndexResultsMetadata(0,0), new ArrayList<IndexResult>(0));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

    /**
	 * Method for mapping uri to label (e.g. for ONKI API).
	 */
	
	// slower than the next method
	/*
	public String getLabel(String uri, String lang) {
		if (this.is.getIndexReader().numDocs() == 0)
			return null;
		//OntologyProcessor processor = OntologyProcessor.getInstance();
		//processor.printRunTime("siblingTime", "56");
		BooleanQuery bq = new BooleanQuery();
		bq.add(new TermQuery(new Term("uri", uri)), BooleanClause.Occur.MUST);
		if (lang != null)
			bq.add(new TermQuery(new Term("language", lang)), BooleanClause.Occur.MUST);
		try {
			//processor.printRunTime("siblingTime", "57");
			TopDocs hits = this.is.search(bq, null, 1);
			//processor.printRunTime("siblingTime", "58");
			// if there is no label for the asked language, get it in some language
			//if (hits.totalHits == 0 && lang != null) {
			//processor.printRunTime("siblingTime", "59");
				//bq = new BooleanQuery();
				//bq.add(new TermQuery(new Term("uri", uri)), BooleanClause.Occur.MUST);
				//hits = this.is.search(bq, null, 1);
			//processor.printRunTime("siblingTime", "60");
			//}
			for (ScoreDoc scoreDoc : hits.scoreDocs) {
			//processor.printRunTime("siblingTime", "61");
				Document doc = this.is.doc(scoreDoc.doc);
			//processor.printRunTime("siblingTime", "62");
				return doc.get("prefLabel");
			}
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
		return null;
	}
	*/
	
	public Map<String, String> getLabels(String uri) {
		if (this.is == null || this.is.getIndexReader().numDocs() == 0)
			return null;
		
		BooleanQuery bq = new BooleanQuery();
		bq.add(new TermQuery(new Term("uri", uri)), BooleanClause.Occur.MUST);
		try {
			Hits hits = this.is.search(bq, new Sort("label"));
			if (hits.length() > 0) {
				HashMap<String, String> labels = new HashMap<String, String>();		
				for (Iterator<Hit> iter = hits.iterator(); iter.hasNext(); ) {
					Document doc = iter.next().getDocument();
					if (!labels.containsKey(doc.get("language")))
						labels.put(doc.get("language"), doc.get("prefLabel"));
				}
				return labels;
			}
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
		return null;
	}
	
	// slower than the previous method
	/*public String getLabel(String uri, List langPreferences) {
		if (this.is == null || this.is.getIndexReader().numDocs() == 0)
			return null;
		
		BooleanQuery bq = new BooleanQuery();
		bq.add(new TermQuery(new Term("uri", uri)), BooleanClause.Occur.MUST);
		
		BooleanQuery lq = new BooleanQuery();
		double weight = 1 / langPreferences.size()*1.0;
		for (int i=0; i<langPreferences.size(); i++) {
			String lang = (String)langPreferences.get(i);
			if (lang != null) {
				TermQuery tq = new TermQuery(new Term("language", lang));
				tq.setBoost((float)weight*(langPreferences.size()-i));
				lq.add(tq, BooleanClause.Occur.SHOULD);
			}
		}
		bq.add(lq, BooleanClause.Occur.MUST);
		
		try {
			Hits hits = this.is.search(bq);
			if (hits.length() > 0) {
				return ((Hit)hits.iterator().next()).getDocument().get("prefLabel");
			}
			//TopDocs hits = this.is.search(bq, null, 1);
			//for (ScoreDoc scoreDoc : hits.scoreDocs) {
			//	return this.is.doc(scoreDoc.doc).get("prefLabel");
			//}
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
		return null;
	}*/
	
	
    /**
	 * Get all the languages of concept labels in the index.
	 */
	
	public Collection<String> getKnownLangs() {
		return this.knownlangs;
	}
	
    /**
	 * Get all the concept types in the index.
	 */
	
	public Collection<String> getKnownTypes() {
		return this.knowntypes;
	}
	
    /**
	 * Get the alphabets in the index.
	 */
	
	public Collection<Character> getKnownAlphabets(String lang) {
		if (lang != null && this.knownalphabets.containsKey(lang))
			return this.knownalphabets.get(lang);
		return this.knownalphabets.get(null);
	}

    /**
	 * Get the label prefixes in the index.
	 */
	public Map<String, Integer> getKnownLabelPrefixes(String lang) {
		if (lang != null && this.knownLabelPrefixes.containsKey(lang))
			return this.knownLabelPrefixes.get(lang);
		return this.knownLabelPrefixes.get(null);
	}
	
    /**
	 * Get all the concept uri namespaces in the index.
	 */
	
	public Collection<String> getKnownNamespaces() {
		return this.knownNamespaces;
	}
	
	@Override
	public String toString() {
		try {
			return "AlphabeticalIndex: indexed items count=" + new IndexSearcher(dir).getIndexReader().numDocs();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
	
    /**
     * Delete index directory.
	 * Deletes all files and subdirectories under dir.
     * Returns true if all deletions were successful.
     * If a deletion fails, the method stops attempting to delete and returns false.
	 */
	 
    private static boolean deleteDir(File dir) {
        if (dir.isDirectory()) {
            String[] children = dir.list();
            for (int i=0; i<children.length; i++) {
                boolean success = deleteDir(new File(dir, children[i]));
                if (!success) {
                    return false;
                }
            }
        }
        // The directory is now empty so delete it
        return dir.delete();
    }
    
    /**
	 * Helper method for main method for creating a list from array of strings.
	 */

	private static List<String> createList(String[] table) {
		List<String> l = new ArrayList<String>();
		for (String s : table)
			l.add(s);
		return l;
	}
    
	private void addLabelIntoKnownLabelPrefixes(String label, String lang) {
		String trimmedLabel = label.trim().toLowerCase();
		if (trimmedLabel.length() > 2)
			trimmedLabel = trimmedLabel.substring(0, 3);
		
		int amount = 1;
		if (this.knownLabelPrefixes.get(lang).containsKey(trimmedLabel))
			amount = this.knownLabelPrefixes.get(lang).get(trimmedLabel)+1;
		this.knownLabelPrefixes.get(lang).put(trimmedLabel, amount);
		
		amount = 1;
		if (this.knownLabelPrefixes.get(null).containsKey(trimmedLabel))
			amount = this.knownLabelPrefixes.get(null).get(trimmedLabel)+1;
		this.knownLabelPrefixes.get(null).put(trimmedLabel, amount);
	}
	
    /**
	 * Main method for testing index.
	 */
    
	public static void main(String[] args) {
		AlphabeticalIndex idx = new AlphabeticalIndex("fi");

		List<IndexItem> testdata = new ArrayList<IndexItem>();
		
		testdata.add(new IndexItem("http://example.org/", "", "auto", "fi", "autot", createList(new String[] {"automobiilit", "henkilöautot"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "auto", "sv", "bilar", createList(new String[] {"personbilar"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "hlo", "fi", "henkilöt", createList(new String[] {"ihmiset"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "hlo", "sv", "personer", createList(new String[] {"personbilar"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "hotelli", "fi", "hotellit", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "hotelli", "sv", "hotell", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "atk", "fi", "ATK", createList(new String[] {"automaattinen tietojenkäsittely"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "ao", "fi", "älykkyysosamäärä", createList(new String[] {"ÄO"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "alykot", "fi", "älykkäät henkilöt", createList(new String[] {"älyköt"}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		/*
		testdata.add(new IndexItem("http://example.org/", "", "S", "fi", "S", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "S100", "fi", "S 100 -proteiinit", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "S-adenosyylihomokysteiini", "fi", "S-adenosyylihomokysteiini", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "S-adenosyylimetioniini", "fi", "S-adenosyylimetioniini", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "Sc", "fi", "S.c.", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "Saab", "fi", "Saab", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "saapas", "fi", "saapas", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		testdata.add(new IndexItem("http://example.org/", "", "Sää", "fi", "Sää", createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), createList(new String[] {}), false, createList(new String[] {})));
		*/
		idx.initialize(testdata);

		System.out.println("querying for: ä@fi");
		for (IndexResult res : idx.search("fi", "ä").getResults())
		//for (IndexResult res : idx.search(null, "s").getResults())
			System.out.println("got: " + res);

		/*System.out.println("Finnish label for <example:autot>: "+idx.getLabel("http://example.org/auto", "fi"));
		System.out.println("Swedish label for <example:autot>: "+idx.getLabel("http://example.org/auto", "sv"));
		System.out.println("English label for <example:autot>: "+idx.getLabel("http://example.org/auto", "en"));
		System.out.println("Null language label for <example:autot>: "+idx.getLabel("http://example.org/auto", null));
		System.out.println("Null language label for unknown uri: "+idx.getLabel("http://example.org/kissa", null));*/
	}
		
	/**
	 * Comparator for sorting the query hits using the ICU Collator.
	 *
	 */
	
	private class IndexResultComparator implements Comparator<IndexResult> {
		
		private String lang;
		
		public IndexResultComparator(String lang) {
			this.lang = lang;
		}
				
		public int compare(IndexResult res1, IndexResult res2) {
			com.ibm.icu.text.Collator icuCollator = com.ibm.icu.text.Collator.getInstance(new Locale(this.lang));
			icuCollator.setStrength(java.text.Collator.PRIMARY);
			return icuCollator.compare((res1.getAltLabel() == null) ? res1.getPrefLabel() : res1.getAltLabel(), 
	    			(res2.getAltLabel() == null) ? res2.getPrefLabel() : res2.getAltLabel());
		}
	}
}