package fi.helsinki.cs.seco.onki.logic;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.HashSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Arrays;
import java.util.Locale;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;

import com.hp.hpl.jena.rdf.model.ResourceFactory;

import fi.helsinki.cs.seco.onki.service.OnkiHierarchyRelation;
import fi.helsinki.cs.seco.onki.utils.Encoder;

/**
 * OnkiResources are used for displaying resources in the user interface and they are
 * typically accessed from the templates.
 */

public class OnkiResource implements Comparable<OnkiResource> {

	public static final int ALL_PROPERTIES = 0;
    public static final int LABEL = 1;
    public static final int ALT_LABEL = 2;
    public static final int DESCRIPTION = 3;
    public static final int RELATED_CONCEPT = 4;
    public static final int EQUIVALENT_CONCEPT = 5;
    
    private Resource resource;
    private List langPreferences;
    
    private static OntologyProcessor processor;
    
    // tree variables
    private Set<OnkiResource> roots = null;
    private List<OnkiResource> parents = null;
    private List<OnkiResource> childs = new LinkedList<OnkiResource>();
    private List<OnkiResource> sortedChilds = null;
    private List<OnkiResource> subnodes = new LinkedList<OnkiResource>();
    
    public static OnkiResource createOnkiResource(String uri, List langPreferences) {
    	if (OnkiResource.processor != null)
    		return OnkiResource.processor.createOnkiResource(uri, langPreferences);
    	return null;
    }
    
    public OnkiResource(Resource res, List langPreferences) {
    	this.setResource(res);
    	this.setLangPreferences(langPreferences);
    }
    
    /**
	 * Set the the ontology processor.
	 */
    
    public static void initialize(OntologyProcessor processor) {
        OnkiResource.processor = processor;
    }

    /**
	 * Get the list of preferred languages for labels etc.
	 */
    
    public List getLangPreferences() {
    	return this.langPreferences;
    }
    
    public void setLangPreferences(List langPreferences) {
        if (langPreferences != null)
        	this.langPreferences = langPreferences;
        else
        	this.langPreferences = new LinkedList();
    }
    
    /**
	 * Get the associated RDF model.
	 */
    
    public Model getModel() {
    	return this.resource.getModel();
    }
    
    /**
	 * Get the underlying RDF resource.
	 */
    
    public Resource getResource() {
        return this.resource;
    }
    
    /**
	 * Set the underlying RDF resource.
	 */
    
    public void setResource(Resource resource) {
        this.resource = resource;
    }
    
    // Concept hierarchy accessing methods ------------------------------------
    
    /**
	 * Set roots of the concept hierarchy.
	 */
    
    public void addRoots(Set<OnkiResource> roots) {
    	if (roots != null) {
    		if (this.roots == null)
    			this.roots = new HashSet<OnkiResource>();
    		this.roots.addAll(roots);
    	}
    }

    /**
	 * Get roots of the concept hierarchy.
	 */
    
    public Set<OnkiResource> getRoots() {
    	if (this.roots == null && this.parents == null)
    		//this.getParents();
        	processor.addParentConcepts(this, this.langPreferences);
    	
    	if (this.roots == null)
    		return new HashSet<OnkiResource>();
    	else
    		return this.roots;
    }
    
    /**
     * Get the subnodes (recursive child nodes) of this node. 
     */
    
    public Collection<OnkiResource> getSubnodes() {
    	return new LinkedList<OnkiResource>(this.subnodes);
    }
    
    /**
	 * Does the node have a specific node as a subnode.
	 */
    
    public boolean hasSubnode(OnkiResource treeId) {    	
    	return this.getSubnodes().contains(treeId);
    }
    
    /**
	 * Does the node have specific nodes as subnodes.
	 */
    
    public boolean hasAllSubnodes(Collection<OnkiResource> nodes) {
    	for (OnkiResource node : nodes)
    		if (!node.equals(this) && !this.getSubnodes().contains(node))
    			return false;
    	return true;
    }
    
    /**
	 * Add a subnode.
	 */
    
    public void addSubnode(OnkiResource node) {
    	if (node != null && !this.subnodes.contains(node) && !node.equals(this))
    		this.subnodes.add(node);
    }
    
    /**
     * Add subnodes.
     */

    public void addSubnodes(Collection<OnkiResource> nodes) {
    	if (nodes != null)
    		for (OnkiResource n: nodes) {
    			if (!n.equals(this))
    				this.addSubnode(n);
    		}
    }
    
    /**
	 * Add a subnode, add it also to the parent nodes of the current node.
	 */
    
    public void addSubnodeRecursively(OnkiResource node) {
    	List<OnkiResource> list = new LinkedList<OnkiResource>();
    	list.add(node);
    	this.addSubnodesRecursively(list);
    }
    	
    /**
	 * Add subnodes, add them also to the parent nodes of the current node.
	 */
    
    public void addSubnodesRecursively(Collection<OnkiResource> nodes) {
    	if (nodes != null && nodes.size()>0) {
    		if (this.hasAllSubnodes(nodes))
    			return;
    		this.addSubnodes(nodes);
    		for(OnkiResource parent: this.getParents()) {
    			if (!parent.equals(this)) {
    				parent.addSubnodesRecursively(nodes);
    			}
    		}
    	}
    }
    
    /**
	 * Add a parent node.
	 */
    
    public void addParent(OnkiResource ores) {
    	if (this.parents == null && ores != null)
    		this.parents = new LinkedList<OnkiResource>();
    	if (ores != null && !this.parents.contains(ores) && !ores.equals(this)) {
    		// If MultiOnki is used, do not add a node which equivalent node is already added.
        	if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
        		boolean hasEquivParent = false;
        		
        		for (Resource equivRes : ores.getEquivalentClassesRes()) {
        			if (this.parents.contains(new OnkiResource(equivRes, this.langPreferences))) {
        				hasEquivParent = true;
        				break;
        			}
        		}

        		if (!hasEquivParent && !this.getEquivalentClassesRes().contains(ores.getResource()))
        			this.parents.add(ores);
        		
        		if (!ores.hasChild(this))
        			ores.addChild(this);
    		}
    		else {
    			this.processor.printRunTime("siblingTime", "17");
    			this.parents.add(ores);
    			this.processor.printRunTime("siblingTime", "18");
        		if (!ores.hasChild(this))
        			ores.addChild(this);
        		this.processor.printRunTime("siblingTime", "19");
    		}
    	}
    }
    
    /**
	 * Does the node have a specific node as a parent node.
	 */
    
    public boolean hasParent(OnkiResource ores) {
    	this.processor.printRunTime("siblingTime", "63");
    	boolean returnValue = this.getParents().contains(ores);
    	this.processor.printRunTime("siblingTime", "64");
    	return returnValue;
    	//return this.getParents().contains(ores);
    }
    
    /**
	 * Get the parent nodes of this node.
	 */
    
    public Collection<OnkiResource> getParents() {
    	if (this.parents == null)
    		processor.addParentConcepts(this, this.langPreferences, 1);
    	
    	if (this.parents == null)
    		return new LinkedList<OnkiResource>();
    	
    	return this.parents;
    }
    
    /**
	 * Get the parent nodes of this node recursively.
	 */
    
    public Collection<OnkiResource> getAllParents() {
    	Set<OnkiResource> parents = new HashSet<OnkiResource>();
    	for (OnkiTreeResource treeres: this.getParentTree()) {
    		parents.add(treeres.getOnkiResource());
    	}
    	return parents;
    }

    /**
	 * Add a child node.
	 */
    
    public void addChild(OnkiResource ores) {
    	if (ores != null && !this.childs.contains(ores) && !ores.equals(this)) {
    		// If MultiOnki is used, do not add a node which equivalent node is already added.
    		if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
        		boolean hasEquivChild = false;
        		
        		for (Resource equivRes : ores.getEquivalentClassesRes()) {
        			if (this.childs.contains(new OnkiResource(equivRes, this.langPreferences))) {
        				hasEquivChild = true;
        				break;
        			}
        		}

        		if (!hasEquivChild && !this.getEquivalentClassesRes().contains(ores.getResource()))
        			this.childs.add(ores);
        		
    	    	this.addSubnodes(ores.getSubnodes());
    	    	this.addSubnode(ores);
        	}
    		else {
    			this.processor.printRunTime("siblingTime", "12");
    			this.childs.add(ores);
    			this.processor.printRunTime("siblingTime", "13");
    			this.addSubnodes(ores.getSubnodes());
    			this.processor.printRunTime("siblingTime", "14");
    	    	this.addSubnode(ores);
    	    	this.processor.printRunTime("siblingTime", "15");
    	    	if (!ores.hasParent(this)) {
        	    	this.processor.printRunTime("siblingTime", "55");	
    	    		ores.addParent(this);
    	    	}
    	    	this.processor.printRunTime("siblingTime", "16");
    		}
    	}
    }
    
    /**
	 * Does the node have a specific node as a child node.
	 */
    
    public boolean hasChild(OnkiResource ores) {
    	return this.childs.contains(ores);
    }
    
    /**
	 * Does the node have child nodes.
	 */
    
    public boolean hasChilds() {
    	return this.amountOfChilds()>0;
    }
    
    /**
	 * Get the amount of child nodes.
	 */
    
    public int amountOfChilds() {
    	return this.childs.size();
    }
    
    /**
     * Get a parent tree of the current node. 
     */
    
    public Collection<OnkiTreeResource> getParentTree() {
    	List<OnkiTreeResource> res = new LinkedList<OnkiTreeResource>();
    	for(OnkiResource r: this.getRoots())
    		res.addAll(r.getSubtrees(this));
    	return res;
    }
    
    /**
     * Get those subtrees of the current node that contain a specific node. 
     */
    
    protected Collection<OnkiTreeResource> getSubtrees(OnkiResource endAt) {
    	return this.getSubtrees(endAt, null, new ArrayList<Boolean>());
    }
    
    protected Collection<OnkiTreeResource> getSubtrees(OnkiResource endAt, OnkiResource parent, List<Boolean> treeindent) {
    	return getSubtrees(endAt, parent, treeindent, new LinkedList<OnkiResource>());
    }
    
    protected Collection<OnkiTreeResource> getSubtrees(OnkiResource endAt, OnkiResource parent, List<Boolean> treeindent, List<OnkiResource> visitednodes) {
    	List<OnkiTreeResource> res = new LinkedList<OnkiTreeResource>();

		if (visitednodes.contains(this))
    		return res;

		// If MultiOnki is used, do not add redundant multihierarchy branches.
    	if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)
    		&& this.hasOneOfListAsEquivalentClass(visitednodes)) {
    		return res;
    	}
    	
    	res.add(new OnkiTreeResource(this, parent, treeindent));

    	if (this.equals(endAt))
    		return res;

    	Collection<OnkiResource> childs = new LinkedList<OnkiResource>();

    	for (OnkiResource c: this.getChilds()) {

    		if (c.hasSubnode(endAt) || c.equals(endAt)) {
    			// If MultiOnki is used, do not add redundant multihierarchy branches.
        		// e.g. mao:aurinkokellot subClassOf "aikaa mittaavat laitteet" subClassOf "kellot" and
        		//      mao:aurinkokellot subClassOf "kellot"
        		//      --> show only the deeper hierarchy
    	    	if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false)) {
	    			boolean redundant = false;
	    			
	    			if (visitednodes.contains(c) || c.hasOneOfListAsEquivalentClass(visitednodes))
	    				redundant = true;
	    			else {
		    			for (OnkiResource c2 : this.getChilds()) {
							if (!visitednodes.contains(c2) && !c2.hasOneOfListAsEquivalentClass(visitednodes)) {
								if (c2.hasChild(c)) {
									redundant = true;
									break;
								}
								else
									for (Resource equivC : c.getEquivalentClassesRes())
										if (c2.hasChild(new OnkiResource(equivC, this.langPreferences))) {
											redundant = true;
											break;
										}
		    				}
		    			}

		    			// added to show hierarchy properly for e.g. yso:laivat
		    			if (endAt.getEquivalentClassesRes().contains(c.getResource()))
		    				c = endAt;
		    			
		    			if (!redundant && !c.equals(endAt)) {
		    				LinkedList<OnkiResource> visitednodes2 = new LinkedList<OnkiResource>(visitednodes);
		    				visitednodes2.add(this);
		    				if (!c.hasSubnodeWithoutRedundancy(endAt, visitednodes2))
		    					redundant = true;
		    			}
		    			
	    			}
					if (!redundant)
	    				childs.add(c);
    			}
    			else
    				childs.add(c);
    		}
    	}

		List<OnkiResource> visitednodes2 = new LinkedList<OnkiResource>(visitednodes);
		visitednodes2.add(this);

    	int childcount = 0;
    	for (OnkiResource c: childs) {
        	List<Boolean> treeindent2 = new ArrayList<Boolean>(treeindent);
        	Boolean hasSiblings = new Boolean(false);
       		if (childs.size()>1 && childcount+1<childs.size())
        		hasSiblings = new Boolean(true);
        	treeindent2.add(hasSiblings);
       		res.addAll(c.getSubtrees(endAt, this, treeindent2, visitednodes2));
    		childcount++;
    	}
    	return res;
    }
    
    /**
	 * Does the current node already have a specific node as a subnode. 
	 */
    
    private boolean hasSubnodeWithoutRedundancy(OnkiResource subnode, List<OnkiResource> visitednodes) {
    	if (visitednodes.contains(this) || this.hasOneOfListAsEquivalentClass(visitednodes))
    		return false;
    	visitednodes.add(this);
    	
    	if (this.hasChild(subnode))
    		return true;
    	
    	//TODO: make a method of this?
    	for (Resource equivRes : subnode.getEquivalentClassesRes())
    		if (this.hasChild(new OnkiResource(equivRes, this.langPreferences)))
    			return true;
    	
		for (OnkiResource child : this.getChilds(subnode))
			if (child.hasSubnodeWithoutRedundancy(subnode, visitednodes))
				return true;
		
		return false;
    }

    /**
	 * Is the current node equivalent with a node in the list. 
	 */
    
    private boolean hasOneOfListAsEquivalentClass(List<OnkiResource> visitednodes) {
    	Collection<Resource> equivClasses = this.getEquivalentClassesRes();
    	for (OnkiResource ores : visitednodes)
    		if (equivClasses.contains(ores.getResource()))
    			return true;
    	return false;
    }
    
    /**
	 * Get those child nodes of this node.
	 */
    
    public Collection<OnkiResource> getChilds() {
    	return this.getChilds(null);
    }

    /**
	 * Get those child nodes of this node that have a specific node as a subnode.
	 */
    
    private Collection<OnkiResource> getChilds(OnkiResource subnode) {
    	List<OnkiResource> childs = new LinkedList<OnkiResource>();
    	processor.printRunTime("siblingTime", "30");
    	this.sortChilds();
    	processor.printRunTime("siblingTime", "31");
    	if (this.sortedChilds == null)
    		return new LinkedList<OnkiResource>();

    	if (subnode == null) {
    		childs = this.sortedChilds;
    		processor.printRunTime("siblingTime", "32");	
    	}
    	else {
	    	for (OnkiResource child: this.sortedChilds) {
	    		processor.printRunTime("siblingTime", "33");	
	    		if (child.hasSubnode(subnode)) {
	    			processor.printRunTime("siblingTime", "34");
	    			childs.add(child);
	    		}
	    	}
    	}
    	processor.printRunTime("siblingTime", "35");
    	return childs;
    }
    
    /**
	 * Sort the child nodes by their label.
	 */
    
    private void sortChilds() {
    	if (this.sortedChilds == null) {
    		if (this.childs == null) {
    			processor.printRunTime("siblingTime", "36");
    			processor.addChildConcepts(this, this.langPreferences);
    			processor.printRunTime("siblingTime", "37");
    		}
    		if (this.childs != null) {
    			this.sortedChilds = this.childs;
    			processor.printRunTime("siblingTime", "38");
    			Collections.sort(this.sortedChilds);
    			processor.printRunTime("siblingTime", "39");
    		}
    	}
    }
    
    /**
	 * Get subnodes of the current node, with indent information.
	 */
    
	public void getSubnodesWithIndent(int currentIndent, int depth, Collection<OnkiHierarchyRelation> relations) {
		if (depth != 0) {
			depth--;
			if (currentIndent >= 0)
				currentIndent++;
			for (OnkiResource child : this.getChilds()) {
				processor.addChildConcepts(child, this.getLangPreferences()); // is inefficient if concept has lots of subconcepts 
				relations.add(new OnkiHierarchyRelation(child.getUri(), this.getUri(), child.getLabel(), currentIndent));
				child.getSubnodesWithIndent(currentIndent, depth, relations);
			}
		}
	}
    
    /**
	 * Get the amount of sibling nodes with a specific parent node.
	 */
	
    public int amountOfSiblingsWithParent(OnkiResource parent) {
    	return this.getSiblingsWithParent(parent).size();
    }
    
    /**
	 * Get the sibling nodes with a specific parent node.
	 */
    
    public Collection<OnkiResource> getSiblingsWithParent(OnkiResource parent) {
    	Collection<OnkiResource> list = new LinkedList<OnkiResource>();

    	processor.printRunTime("siblingTime", "25");    	
    	if (!this.hasParent(parent)) {
    		processor.printRunTime("siblingTime", "26");
        	if (processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_ENABLED) && !processor.getConfig().getBoolean(OnkiConfigurator.ELEM_MULTIONKI_NOMAPPINGSINHIERARCHY, false))
    			parent = this.getEquivParent(parent);
    	}
    	
    	processor.printRunTime("siblingTime", "27");
    	if (!this.hasParent(parent))
    		return list;
    	
    	processor.printRunTime("siblingTime", "28");
    	list.addAll(parent.getChilds());
    	processor.printRunTime("siblingTime", "29");
    	list.remove(this);
    	return list;
    }
    
    /**
	 * Get a node that is equivalent to a specifig parent node and is also a parent node 
	 * of the current node. 
	 */
    
    public OnkiResource getEquivParent(OnkiResource parent) {
		for (Resource equivParentRes : parent.getEquivalentClassesRes())
			for (OnkiResource parent2 : this.getParents())
				if (parent2.getUri().equals(equivParentRes.getURI()))
					return parent2;
    	return parent;
    }
    
    // methods for accessing the content -------------------------
    
    /**
	 * Get the types of the resource.
	 */
    
    public Collection<OnkiResource> getTypes() {
    	LinkedList<OnkiResource> types = new LinkedList<OnkiResource>();
    	for (StmtIterator stmtI = this.resource.listProperties(RDF.type); stmtI.hasNext(); ) {
    		Statement stmt = stmtI.nextStatement();
    		this.getTypesRecursively((Resource)stmt.getObject(), types);
    	}
    	return types;
    }
    
    /**
	 * Get the types of the resource transitively (superclasses of asserted types).
	 */
    
    public void getTypesRecursively(Resource type, LinkedList<OnkiResource> types) {
    	OnkiResource typeRes = processor.createOnkiResource(type.getURI(), langPreferences);
    	if (!types.contains(typeRes)) {
    		types.add(typeRes);
	    	for (StmtIterator superStmtI = type.listProperties(RDFS.subClassOf); superStmtI.hasNext(); ) {
	    		Resource superType = (Resource)superStmtI.nextStatement().getObject();
				this.getTypesRecursively(superType, types);
			}
    	}
    }
    
    /**
	 * Does the resource types as a specific type.
	 */
    
    public boolean hasType(String url) {
    	return this.getTypes().contains(createOnkiResource(url, this.langPreferences));
    }
    
    /**
	 * Does the resource represent a group (of concepts).
	 */
    
    public boolean isGroup() {
    	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED)) {
    		RDFNode valueNode = null;
    		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_IS_GROUP_VALUE)) {
    			String value = processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_IS_GROUP_VALUE);
    			if (processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_IS_GROUP_VALUE_TYPE).equals("resource"))
    				valueNode = this.getModel().getResource(value);
    			else // literal
    				valueNode = this.getModel().createLiteral(value);
    		}
    		
    		Property isGroup = this.resource.getModel().createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_IS_GROUP_PROPERTY));
    		StmtIterator iter;
    		if (processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_IS_GROUP_PROPERTY_TYPE).equals("normal"))
    			iter = this.resource.getModel().listStatements(this.resource, isGroup, valueNode);
    		else // inverse
    			iter = this.resource.getModel().listStatements((Resource)valueNode, isGroup, this.resource);
    		if (iter.hasNext())
    			return true;
    	}
   		return false;    	
    }
    
    /**
	 * Get the groups the concept belongs to.
	 */
    
    public Collection<OnkiResource> getGroups() {
    	// TODO: utilize Lucene index?
    	List<OnkiResource> groups = new LinkedList<OnkiResource>();
    	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_GROUP_ENABLED)) {
    		Property belongs = this.resource.getModel().createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_BELONGS_TO_GROUP_PROPERTY));
    		if (processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_GROUP_BELONGS_TO_GROUP_PROPERTY_TYPE).equals("normal"))
    			for (StmtIterator iter = this.resource.listProperties(belongs); iter.hasNext(); ) {
    				OnkiResource ores = processor.createOnkiResource(iter.nextStatement().getResource().getURI(), langPreferences);
    				if (ores.isGroup())
    					groups.add(ores);
    			}
    		else // inverse
    			for (ResIterator iter = this.resource.getModel().listSubjectsWithProperty(belongs, this.resource); iter.hasNext(); ) {
    				OnkiResource ores = processor.createOnkiResource(iter.nextResource().getURI(), langPreferences);
    				if (ores.isGroup())
    					groups.add(ores);
    			}
    	}
    	return groups;
    }

    /**
	 * Get the query expansions of the concept.
	 */
    
    public Collection<OnkiResource> getQueryExpansions() {
    	return this.getQueryExpansions(null);
    }
    
    public Collection<OnkiResource> getQueryExpansions(LinkedList<OnkiResource> alreadyProcessed) {
    	if (alreadyProcessed == null)
    		alreadyProcessed = new LinkedList<OnkiResource>();
   		alreadyProcessed.add(this);

    	// TODO: not efficient to access conf file multiple times
    	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_ENABLED) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_ENABLED) && processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY)) {
    		Property prop = this.resource.getModel().createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY));
    		boolean inverse = false;
    		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_INVERSE) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_INVERSE))
    			inverse = true;
    		boolean symmetric = false;
    		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_SYMMETRIC) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_SYMMETRIC))
    			symmetric = true;
    		boolean transitive = false;
    		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_TRANSITIVE) && processor.getConfig().getBoolean(OnkiConfigurator.ELEM_CLASSES_QUERY_EXPANSION_PROPERTY_TRANSITIVE))
    			transitive = true;
    		
    		// TODO: case "symmetric & transitive" leads to duplicate values 
    		if (symmetric || !inverse) {
    			for (StmtIterator iter = this.resource.listProperties(prop); iter.hasNext(); ) {
    				OnkiResource ores = processor.createOnkiResource(iter.nextStatement().getResource().getURI(), langPreferences);
    				if (!alreadyProcessed.contains(ores)) {
	    				if (transitive)
	    					ores.getQueryExpansions(alreadyProcessed);
	    				else
	    					alreadyProcessed.add(ores);
    				}
    			}
    		}
    		if (symmetric || inverse) {
    			for (StmtIterator iter = this.resource.getModel().listStatements(null, prop, this.getResource()); iter.hasNext(); ) {
    				OnkiResource ores = processor.createOnkiResource(iter.nextStatement().getSubject().getURI(), langPreferences);
    				if (!alreadyProcessed.contains(ores)) {
	    				if (transitive)
	    					ores.getQueryExpansions(alreadyProcessed);
	    				else
	    					alreadyProcessed.add(ores);
    				}
    			}
    		}
    	}
    	return alreadyProcessed;
    }
    
    /**
	 * Get the properties of the concept.
	 */
    
    public Collection<OnkiStatement> getProperties() {
    	return this.getProperties(true, null, false, false);
    }
    
    public Collection<OnkiStatement> getProperties(Set<Property> properties) {
    	return this.getProperties(true, properties, false, false);
    }
    
    public Collection<OnkiStatement> getProperties(boolean undirect, Set<Property> properties, boolean sortedAndFilteredForOnkiFrontend, boolean filtered) {
    	List<IPropertyFiddler> fiddlerInstances = null;
    	if (undirect) {
	    	String[] fiddlers = processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_PROPERTY_FIDDLERS_FIDDLER);
	    	fiddlerInstances = new LinkedList<IPropertyFiddler>();
	    	for (int i=0; i<fiddlers.length; i++) {
	    		try {
	    			fiddlerInstances.add((IPropertyFiddler) java.lang.Class.forName(fiddlers[i]).newInstance());
	    		} catch(Exception e) {
	    			System.err.println("Exception catched in getProperties when instantiating property fiddlers: "+e);
	    		}
	    	}
    	}
    	
    	// hide properties configured as hidden only if all properties are asked (for property listing in UI)
    	List<String> hiddenProps = new ArrayList<String>();
    	if (properties == null) {
    		String[] hiddenPropsArray = processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_HIDDEN_PROPERTY);
    		hiddenProps = Arrays.asList(hiddenPropsArray);
    	}
    	
    	String[] additionalPropsArray = processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_ADDITIONAL);
    	List<String> additionalProps = Arrays.asList(additionalPropsArray);
    	
    	String[] urlPropsArray = processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_URL_PROPERTIES_PROPERTY);
    	List<String> urlProps = Arrays.asList(urlPropsArray);
    	
    	List<OnkiStatement> result = new LinkedList<OnkiStatement>();
    	for (StmtIterator stmtI = this.resource.listProperties(); stmtI.hasNext(); ) {
    		Statement stmt = stmtI.nextStatement();
    		
    		// if propertyFiltering is in use, choose only those properties that match the filter
    		if (properties != null && !properties.contains(stmt.getPredicate()))
    			continue;
    		
    		// search for matching property fiddlers
    		List<OnkiStatement> fiddlerResults = null;

    		if (fiddlerInstances != null) {
	    		for (IPropertyFiddler f: fiddlerInstances) {
	    			List<OnkiStatement> fiddlerResults2 = f.findForward(this, stmt);
	    			if (fiddlerResults2 != null) {
	    				if (fiddlerResults == null)
	    					fiddlerResults = new LinkedList<OnkiStatement>();
	    				fiddlerResults.addAll(fiddlerResults2);
	    			}
	    		}
    		}
		
    		if (fiddlerResults == null)
    			result.add(new OnkiStatement(this, stmt.getPredicate(), stmt.getObject(), this.langPreferences));
    		else
    			result.addAll(fiddlerResults);

    	}

    	// find for backward links
		if (fiddlerInstances != null) {
    		for (IPropertyFiddler f: fiddlerInstances) {
    			List<OnkiStatement> fiddlerResultsBackward = f.findBackward(this);
    			if (fiddlerResultsBackward != null) {
					for (OnkiStatement stmt : fiddlerResultsBackward) {
						if (properties != null && !properties.contains(this.getModel().getProperty(stmt.getPredicate().getUri())))
							continue;
						result.add(stmt);
					}
    			}
    		}
		}

		// remove hidden statements and add additional types to statements
		// do it here to avoid duplicated code above		
		
		List<OnkiStatement> finalresult = new LinkedList<OnkiStatement>();
		
		for (OnkiStatement os: result) {
			if (hiddenProps.contains(os.getPredicate().getUri()))
				continue;
			if (additionalProps.contains(os.getPredicate().getUri()))
				os.setAdditional(true);
			if (urlProps.contains(os.getPredicate().getUri()))
				os.setObjectAsURL(true);
			finalresult.add(os);
		}

		// TODO: constants of this class should be used, e.g. OnkiResource.LABEL, Jouni 30.6.2009
		if (sortedAndFilteredForOnkiFrontend) {
			List<OnkiStatement> preorderedStatements = new LinkedList<OnkiStatement>();
			List<OnkiStatement> removedStatements = new LinkedList<OnkiStatement>();
			List<String> propertyOrder = Arrays.asList(processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_REST_API_PROPERTY_ORDER_PROPERTY));
			
			if (filtered) {
				String[] hideProps = {OnkiConfigurator.ELEM_CLASSES_LABEL,
										OnkiConfigurator.ELEM_CLASSES_INDEX_ALT_LABEL,
										OnkiConfigurator.ELEM_CLASSES_FOLLOW,
										OnkiConfigurator.ELEM_CLASSES_EQUIVALENT_CLASS_PROPERTIES_PROPERTY,
										OnkiConfigurator.ELEM_CLASSES_REST_API_HIDDEN_PROPERTY};
				for (String propSet : hideProps) {
					if (processor.getConfig().containsKey(propSet)) {
						for (String prop : processor.getConfig().getStringArray(propSet)) {
							if (!propertyOrder.contains(prop)) {
								for (OnkiStatement os : finalresult) {
									if (os.getPredicate().getUri().equals(prop) && !removedStatements.contains(os)) {
										removedStatements.add(os);
									}
								}
							}
						}
					}
				}
			}
			
			finalresult.removeAll(removedStatements);
			Collections.sort(finalresult);

			for (String prop : propertyOrder) {
				for (OnkiStatement os : finalresult) {
					if (os.getPredicate().getUri().equals(prop) && !preorderedStatements.contains(os)) {
						preorderedStatements.add(os);
					}
				}
			}
			
			finalresult.removeAll(preorderedStatements);
			finalresult.addAll(0, preorderedStatements);
		}
		
		else
			Collections.sort(finalresult);

    	return finalresult;
    }
    
    /**
	 * Get the label of the concept in the default language.
	 */
    
    public String getLabel() {
    	return this.getLabel(null);
    }
    
    /**
	 * Get the label of the concept.
	 */
    
    public String getLabel(String lang) {
    	List langPrefences = this.langPreferences;
    	if (lang != null)
			((LinkedList)langPreferences).addFirst(lang);
    	return processor.getConceptLabel(this.getUri(), langPreferences);
    }
    
    /*public String getLabel(String lang) {
    	if (!this.resourceAvailable())
    		return "(no resource)";
    	
		// get label from lucene
		String labelFromLucene = null;
		processor.printRunTime("siblingTime", "52");
		List langPreferences = this.langPreferences;
		if (lang != null)
			langPreferences.add(0, lang);
		
		for (Iterator langI = langPreferences.iterator(); langI.hasNext(); ) {
			labelFromLucene = this.processor.getConceptLabel(this.getUri(), (String)langI.next());
			if (labelFromLucene != null) {
				processor.printRunTime("siblingTime", "53.1");
				return labelFromLucene;
			}
			processor.printRunTime("siblingTime", "52.1");
		}
		processor.printRunTime("siblingTime", "53");
		
		// label was not found from lucene (e.g. properties), use jena instead   	
        for (Iterator langI = langPreferences.iterator(); langI.hasNext(); ) {
        	String currentLang = (String)langI.next();
		    for (Iterator iter = processor.getLabelProps().iterator();  iter.hasNext();) {
		        Property p = (Property) iter.next();
		        try {
		        	if (processor.getConcept(resource.getURI()) != null) {
//		        		processor.printRunTime("siblingTime", "46");
				        for (StmtIterator stmtIter = processor.getConcept(resource.getURI()).listProperties(p); stmtIter.hasNext();) {
//				        	processor.printRunTime("siblingTime", "47");
				            Statement stmt = stmtIter.nextStatement();
//				            processor.printRunTime("siblingTime", "50");
				            if (stmt == null) continue;
		
				            if (stmt.getLanguage() != null && stmt.getLanguage().equals(currentLang) ||
				            	currentLang == null) {
				            	processor.printRunTime("siblingTime", "51");
				            	
				            	return stmt.getString();
				        	}
//				            processor.printRunTime("siblingTime", "48");
				        }
		        	}
		        } catch(HasNoModelException e) {} // Occurs when the property is not connected with a model.
		    }
        }
        processor.printRunTime("siblingTime", "54");

        return resource.getLocalName()+"*";
    }*/

    /**
	 * Get the label of the concept escaped for XHTML representation.
	 */
    
    public String getLabelEscaped() {
		// OnkiResource.getLabel() can be null if the resource is part of a OnkiStament
    	// created with a PropertyFiddler and the resource is not present in the RDF model
    	if (getLabel() == null)
    		return "*";
    	return Encoder.escapeXml(getLabel());
    }
    
    public String getLabelEscaped(String lang) {
		// OnkiResource.getLabel() can be null if the resource is part of a OnkiStament
    	// created with a PropertyFiddler and the resource is not present in the RDF model
    	if (getLabel(lang) == null)
    		return "*";
    	return Encoder.escapeXml(getLabel(lang));	
    }
    
    /**
	 * Is the underlying resource set.
	 */
    
    private boolean resourceAvailable() {
    	return this.resource != null;
    }
    
    /**
	 * Get certain properties of the concept.
	 */
    
    public Collection<OnkiStatement> getLabels(String lang, String propertyLang) {
    	return this.getProperties(OnkiResource.LABEL, lang, propertyLang);
    }
    
    public Collection<OnkiStatement> getAltLabels(String lang, String propertyLang) {
    	return this.getProperties(OnkiResource.ALT_LABEL, lang, propertyLang);
    }
    
    public Collection<OnkiStatement> getDescriptions(String lang, String propertyLang) {
    	return this.getProperties(OnkiResource.DESCRIPTION, lang, propertyLang);
    }
    
    public Collection<OnkiStatement> getRelatedConcepts(String lang) {
    	return this.getProperties(OnkiResource.RELATED_CONCEPT, null, lang);
    }
    
    public Collection<OnkiStatement> getEquivalentConcepts(String lang) {
        // TODO: use getEquivalentClassesRes() ? (adds support for transitivity and symmetry)
    	return this.getProperties(OnkiResource.EQUIVALENT_CONCEPT, null, lang);
    }
    
    public Collection<OnkiStatement> getProperties(int propertyType, String lang, String propertyLang) {
        // TODO: this should use OnkiResource.getProperties(Set<Property)
    	if (!this.resourceAvailable())
    		return null;
    	LinkedList<OnkiStatement> stmts = new LinkedList<OnkiStatement>();

    	List<String> stmtLangPreferences = this.langPreferences;
    	if (propertyLang != null) {
    		if (stmtLangPreferences == null)
    			stmtLangPreferences = new LinkedList<String>();
    		((LinkedList)stmtLangPreferences).addFirst(propertyLang);
    	}

    	List<String> props = null;
    	if (processor.getProps(propertyType) != null)
    		props = Arrays.asList(processor.getProps(propertyType));
    	
    	if (props != null && processor.getConcept(resource.getURI()) != null) {
    		Set<Property> properties = new HashSet<Property>();
    		for (String prop : props)
    			properties.add(this.getModel().getProperty(prop));
    		for (OnkiStatement stmt : this.getProperties(properties)) {
    			if (stmt == null) continue;
				// TODO: lang filtering (lang != null) works only for literal objects, resource objects cause exception -> error handling
				if (lang == null || stmt.getLang().equals(lang)) {
					stmts.add(new OnkiStatement(this, this.getModel().getProperty(stmt.getPredicate().getUri()), stmt.getObject(), stmtLangPreferences));
				}
    		}
    	}
    	
    	return stmts;
    }
    
    /**
	 * Get the uri of the concept.
	 */
    
    public String getUri() {
        if (this.resource != null) {
            return this.resource.getURI();
        } else {
            return "";
        }
    }
    
    /**
	 * Get the namespace of the concept URI.
	 */
    
    public String getNamespace() {
        return resource.getNameSpace();
    }
    
    /**
	 * Get the namespace prefix of the concept URI.
	 */
    
    public String getNamespacePrefix() {
        String ns = resource.getModel().getNsURIPrefix(resource.getNameSpace());
        if (ns == null)
            ns = "";
        return ns;
    }
    
    /**
	 * Get a property of the underlying resource.
	 */
    
    public Statement getProperty(Property prop) {
        return resource.getProperty(prop);
    }
    
    public StmtIterator getProperties(Property prop) {
    	return resource.listProperties(prop);
    }
    
    /**
	 * Has the concept equivalent concepts.
	 */
    
    public boolean hasEquivalentClass() {
    	return processor.hasEquivalentClass(this.getResource());
    }
    
    /**
	 * Get the equivalent concepts (for UI).
	 */
    
    public Collection<OnkiResource> getEquivalentClasses() {
    	HashSet<OnkiResource> set = new HashSet<OnkiResource>();
    	for (Resource res : this.getEquivalentClassesRes())
    		set.add(new OnkiResource(res, this.langPreferences));
    	return set;
    }
    
    /**
	 * Get the namespace prefixes of the equivalent concepts (for UI).
	 */
    
    public Collection<String> getEquivalentClassNSPrefixes() {
    	TreeSet<String> set = new TreeSet<String>();
    	for (OnkiResource equivOres : this.getEquivalentClasses())
    		set.add(equivOres.getNamespacePrefix());
    	return set;
    }
    
    /**
	 * Get the equivalent concepts as RDF resources.
	 */
    
    public Collection<Resource> getEquivalentClassesRes() {
    	return processor.getEquivalentClassesRes(this.getResource(), null);
    }
    
    /**
	 * Is the resource a top concept of a SKOS ConceptScheme.
	 */
    
    public boolean isTopConcept() {
    	StmtIterator stmtIter = this.getProperties(RDF.type);
    	while (stmtIter.hasNext())
    		if (stmtIter.nextStatement().getResource().getURI().equals(OntologyProcessor.SKOS_NS+"TopConcept"))
    			return true;
    	return processor.isTopConcept(this.resource);
    }
    
    /**
	 * Is the resource a SKOS ConceptScheme.
	 */
    
    public boolean isConceptScheme() {
    	StmtIterator stmtIter = this.getProperties(RDF.type);
    	while (stmtIter.hasNext()) 
    		if (stmtIter.nextStatement().getResource().getURI().equals(OntologyProcessor.SKOS_NS+"ConceptScheme"))
    			return true;
    	return false;
    }
    
    /**
	 * Is the resource a SKOS Collection.
	 */
    
    public boolean isCollection() {
    	StmtIterator stmtIter = this.getProperties(RDF.type);
    	while (stmtIter.hasNext()) 
    		if (stmtIter.nextStatement().getResource().getURI().equals(OntologyProcessor.SKOS_NS+"Collection"))
    			return true;
    	return false;
    }
    
    /**
	 * Get the weighted mappings between concepts.
	 */
    
    public Map<OnkiResource, Collection<OnkiConceptMapping>> getMappings() {
    	List<IMappingFiddler> fiddlerInstances = null;

    	String[] fiddlers = processor.getConfig().getStringArray(OnkiConfigurator.ELEM_CLASSES_MAPPING_FIDDLERS_FIDDLER);
    	fiddlerInstances = new LinkedList<IMappingFiddler>();
    	for (int i=0; i<fiddlers.length; i++) {
    		try {
    			fiddlerInstances.add((IMappingFiddler) java.lang.Class.forName(fiddlers[i]).newInstance());
    		} catch(Exception e) {
    			System.err.println("Exception catched in getMappings when instantiating mapping fiddlers: "+e);
    		}
    	}
    	
    	Map<OnkiResource, Collection<OnkiConceptMapping>> fiddlerResults = null;

		if (fiddlerInstances != null) {
    		for (IMappingFiddler f: fiddlerInstances) {
    			Map<OnkiResource, Collection<OnkiConceptMapping>> fiddlerResults2 = f.findMappings(this);
    			if (fiddlerResults2 != null) {
    				if (fiddlerResults == null)
    					fiddlerResults = new LinkedHashMap<OnkiResource, Collection<OnkiConceptMapping>>();
    				fiddlerResults.putAll(fiddlerResults2);
    			}
    		}
		}
		return fiddlerResults;
    }
    
    /**
	 * Get the geographical polygon of the concept.
	 */
    
    public String getPolygon() {
    	//TODO: support multiple values
    	if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POLYGON_PROPERTY)) {
    		Statement stmt = this.getProperty(ResourceFactory.createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POLYGON_PROPERTY)));
   			if (stmt != null && stmt.getObject().isLiteral())
   				return stmt.getString();
    	}
    	return null;
    }

    /**
	 * Get the geographical point of the concept.
	 */
    
    public String getPoint() {
    	//TODO: support multiple values
		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY_LAT_PROPERTY) && processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY_LONG_PROPERTY)) {
			Statement latStmt = this.getProperty(ResourceFactory.createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY_LAT_PROPERTY)));
			Statement longStmt = this.getProperty(ResourceFactory.createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY_LONG_PROPERTY)));
			if (latStmt != null && latStmt.getObject().isLiteral() && longStmt != null && longStmt.getObject().isLiteral())
				return latStmt.getString()+","+longStmt.getString();
		}
		if (processor.getConfig().containsKey(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY)) {
			Statement stmt = this.getProperty(ResourceFactory.createProperty(processor.getConfig().getString(OnkiConfigurator.ELEM_CLASSES_MAP_PROPERTIES_POINT_PROPERTY)));
   			if (stmt != null && stmt.getObject().isLiteral())
   				return stmt.getString();
		}
    	return null;
    }

    /**
	 * OnkiResources are equal if their URIs are the same.
	 */
    
    public boolean equals(Object arg0) {
       	 return (arg0 instanceof OnkiResource && this.getUri() != null && ((OnkiResource)arg0).getUri() != null && ((OnkiResource)arg0).getUri().trim().equals(this.getUri().trim()));
    }
    
    public int hashCode() {
    	return resource.hashCode();
    }
    
    /**
	 * OnkiResources are compared by their label.
	 */
    
    public int compareTo(OnkiResource ores) {
    	if (ores == null)
    		return -1;
    	// to handle accents etc. (e.g. "ś" == "s")
    	processor.printRunTime("siblingTime", "40");
    	com.ibm.icu.text.Collator icuCollator;
    	if (this.langPreferences.size() > 0 && this.langPreferences.get(0) != null)
    		icuCollator = com.ibm.icu.text.Collator.getInstance(new Locale((String)this.langPreferences.get(0)));
    	else
    		icuCollator = com.ibm.icu.text.Collator.getInstance();
    	processor.printRunTime("siblingTime", "41");
    	icuCollator.setStrength(java.text.Collator.PRIMARY);
    	processor.printRunTime("siblingTime", "42");
    	//return icuCollator.compare(this.getLabel(), ores.getLabel());
    	String thisLabel = this.getLabel();
    	processor.printRunTime("siblingTime", "44");
    	String oresLabel = ores.getLabel();
    	processor.printRunTime("siblingTime", "45");
    	int returnValue = icuCollator.compare(thisLabel, oresLabel);
    	processor.printRunTime("siblingTime", "43");
    	return returnValue;
    }
    
    // ------------ debug tools ----------------------------
    
    public String toString() {
        return "OnkiResource: " +
        		"resource: " + this.resource.toString();
    }
    
    public String stringInfo() {
    	return this.toString() + "\n" +
        		"subnodes: " + this.subnodes + "\n" +
        		"roots: " + this.roots + "\n" +
        		"parents: " + this.parents + "\n" +
        		"childs: " + this.childs;
    }
    
    public void setRoots(Set<OnkiResource> roots) {
    	this.roots = roots;
    }
    
    public void setParents(List<OnkiResource> parents) {
    	this.parents = parents;
    }
}