package fi.helsinki.cs.seco.onki.logic;

import java.util.List;
import java.util.Locale;

import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.vocabulary.XSD;

import fi.helsinki.cs.seco.onki.utils.Encoder;

/**
 * Represents a statement about OnkiResource.
 */

public class OnkiStatement implements Comparable<OnkiStatement> {

	public static final int TYPE_INDIRECT = 1;
	public static final int TYPE_ADDITIONAL = 2;
	public static final int TYPE_INVERSE = 4;
	public static final int TYPE_OBJECT_AS_URL = 8;
	
	private OnkiResource subject;
	private OnkiResource predicate;
	private RDFNode object;
	private List langPreferences;
	private int type;
	
	public OnkiStatement(OnkiResource subject, Property predicate, RDFNode object, List langPreferences) {
		this(subject, predicate, object, langPreferences, 0);
	}
	
	public OnkiStatement(OnkiResource subject, Property predicate, RDFNode object, List langPreferences, int type) {
		this.langPreferences = langPreferences;
		this.type = type;
		
		this.subject = subject;
		this.predicate = new OnkiResource(predicate, this.langPreferences);
		this.object = object;
	}

    /**
	 * Getters and setters for the flag bits of the type of the statement.
	 */
	
	public boolean isIndirect() {
		return (this.type & OnkiStatement.TYPE_INDIRECT) == OnkiStatement.TYPE_INDIRECT; 
	}

	public boolean isAdditional() {
		return (this.type & OnkiStatement.TYPE_ADDITIONAL) == OnkiStatement.TYPE_ADDITIONAL; 
	}

	public boolean isInverse() {
		return (this.type & OnkiStatement.TYPE_INVERSE) == OnkiStatement.TYPE_INVERSE; 
	}
	
	public boolean isObjectAsURL() {
		return (this.type & OnkiStatement.TYPE_OBJECT_AS_URL) == OnkiStatement.TYPE_OBJECT_AS_URL;
	}
	
	public void setIndirect(boolean b) {
		this.setType(OnkiStatement.TYPE_INDIRECT, b); 
	}

	public void setAdditional(boolean b) {
		this.setType(OnkiStatement.TYPE_ADDITIONAL, b); 
	}

	public void setInverse(boolean b) {
		this.setType(OnkiStatement.TYPE_INVERSE, b); 
	}
	
	public void setObjectAsURL(boolean b) {
		this.setType(OnkiStatement.TYPE_OBJECT_AS_URL, b);
	}
	
	private void setType(int mask, boolean value) {
		if (value)
			this.type = this.type |= mask;
		else
			this.type = this.type & ~mask;
	}
	
    /**
	 * Get the subject of the statement.
	 */
	
	public OnkiResource getSubject() {
		return this.subject;
	}
	
    /**
	 * Get the predicate of the statement.
	 */
	
	public OnkiResource getPredicate() {
		return this.predicate;
	}
	
    /**
	 * Get the language code of the object.
	 */
	
	public String getLang() {
		if (this.object.isLiteral()) {
			Literal lit = (Literal)this.object;
			return lit.getLanguage();
		}
		return "";
	}
	
    /**
	 * Is the language code of the object set.
	 */
	
	public boolean hasLang() {
		return this.object.isLiteral() && ((Literal)this.object).getLanguage() != "";
	}
	
    /**
	 * Get the object of the statement as OnkiResource if possible, otherwise return null.
	 */
	
	public OnkiResource getObjectAsOnkiResource() {
		if (this.objectIsURIResource())
			return OnkiResource.createOnkiResource(this.getObjectAsURI(), this.langPreferences);
		return null;
	}
	
    /**
	 * Get the object of the statement as RDFNode.
	 */
	
	public RDFNode getObject() {
		return this.object;
	}
	
    /**
	 * Is the type of object of the statement XSD:anyUri.
	 */
	
	public boolean objectIsAnyURILiteral() {
		return (this.object.isLiteral() && ((Literal)this.object).getDatatypeURI() != null && ((Literal)this.object).getDatatypeURI().equals(XSD.anyURI.getURI()));
	}
	
    /**
	 * Is the object of the statement URI resource.
	 */
	
	public boolean objectIsURIResource() {
		return this.object.isURIResource();
	}
	
    /**
	 * Get the URI of the object.
	 */
	
	public String getObjectAsURI() {
		if (this.object.isURIResource())
			return ((Resource)this.object).getURI();
		return "";
	}
	
    /**
	 * Get the object of the statement as string.
	 */
	
	public String getObjectAsString() {
		if (this.object.isLiteral())
			return ((Literal)this.object).getString();
		if (this.object.isURIResource())
			return this.getObjectAsURI();
		return "";
	}
	
    /**
	 * Get the object of the statement as string, escaped for XHTML presentation.
	 */
	
	public String getObjectAsStringEscaped() {
		return Encoder.escapeXml(this.getObjectAsString());
	}
	
	/**
	 * For comparing OnkiStatement objects.
	 */
	
	public int compareTo(OnkiStatement os) {
		/*
		 * Comment the following if clause if the "additional" properties are to be sorted
		 * among the other properties (and not after them).
		 * In that case also the template concept-view.vm should be modified.
		 */
		if (os.isAdditional() != this.isAdditional()) {
			if (os.isAdditional() == true)
				return -1;
			else
				return 1;
		}
		
		if (os.hasLang() && this.hasLang())
			if (!this.getLang().trim().equalsIgnoreCase(os.getLang().trim()))
				return this.getLang().compareToIgnoreCase(os.getLang());
		
		if (this.hasLang() && !os.hasLang())
			return -1;
		
		if (os.hasLang() && !this.hasLang())
			return 1;
		
		com.ibm.icu.text.Collator icuCollator;
		if (this.langPreferences.size() > 0 && this.langPreferences.get(0) != null)
			icuCollator = com.ibm.icu.text.Collator.getInstance(new Locale((String)this.langPreferences.get(0)));
		else
			icuCollator = com.ibm.icu.text.Collator.getInstance();
		icuCollator.setStrength(java.text.Collator.PRIMARY);
		
		if (!os.getPredicate().equals(this.getPredicate()))
			// OnkiResource.getLabel() of a predicate can be null if the statement is created with PropertyFiddler
			// and the resource is not present in the RDF model
			if (this.getPredicate().getLabel() != null && os.getPredicate().getLabel() != null
				&& !this.getPredicate().getLabel().equalsIgnoreCase(os.getPredicate().getLabel()))
				return icuCollator.compare(this.getPredicate().getLabel(), os.getPredicate().getLabel());
			else
				return icuCollator.compare(this.getPredicate().getUri(), os.getPredicate().getUri());

		if (this.getObjectAsOnkiResource() != null && os.getObjectAsOnkiResource() != null)
			return icuCollator.compare(this.getObjectAsOnkiResource().getLabel(), os.getObjectAsOnkiResource().getLabel());
		
		return icuCollator.compare(this.getObjectAsString(), os.getObjectAsString());
	}
	
	/**
	 * OnkiStatement objects are equal if their subjects, predicates and objects are equal.
	 */
	
	public boolean equals (Object arg0) {
		return (arg0 instanceof OnkiStatement && this.getSubject().equals(((OnkiStatement)arg0).getSubject())
				&& this.getPredicate().equals(((OnkiStatement)arg0).getPredicate())
				&& this.getObject().equals(((OnkiStatement)arg0).getObject()));
	}
	
	public int hashCode() {		
		int result = this.getSubject().hashCode();
		// Math.exp(Math.log(Integer.MAX_VALUE)/"number of fields") == 1290
		result = 1290 * result + this.getPredicate().hashCode();
		result = 1290 * result + this.getObject().hashCode();
			
		return result;
	}
}