package fi.helsinki.cs.seco.onki.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

/**
 * Class for generating ONKI SKOS browser URLs for links.
 */

public class OnkiURLGenerator {
	private String baseurl;
	private Map<String,String[]> requestParams;
	private String encoding;
	private Map<String,String> defaultParams = new HashMap<String,String>();
	
	public static final String PARAM_CONCEPT = "c";
	public static final String PARAM_LANG = "l";
	public static final String PARAM_PAGE = "p";
	public static final String PARAM_QUERY = "q";
	public static final String PARAM_QUERY_TYPE = PARAM_QUERY+"t"; // only matches res's with this type
	public static final String PARAM_QUERY_PARENT = PARAM_QUERY+"p"; // only matches res's that are under this resource
	public static final String PARAM_QUERY_GROUP = PARAM_QUERY+"g"; // only matches res's belonging to this group
	public static final String PARAM_GROUP_QUERY = "g"+PARAM_QUERY; // only matches res's belonging to this group
	public static final String PARAM_FETCH_CONCEPT_MODE = "fc";
	public static final String PARAM_REFERER = "r";
	public static final String PARAM_RDF = "rdf";
	// TEMPORARY PARAM FOR DEMO, Jouni 15.9.2008
	public static final String PARAM_MAPPINGS = "m";
	
	public OnkiURLGenerator(String baseurl, Map<String,String[]> requestParams, String encoding) {
		this.baseurl = baseurl;
		this.requestParams = requestParams;
		this.encoding = encoding;
	}
	
	public void setDefault(String key, int value) {
		this.setDefault(key, ""+value);
	}
	
	public void setDefault(String key, String value) {
		if (key != null && 
				(
				key.equals(PARAM_PAGE) ||
				key.equals(PARAM_LANG) ||
				key.equals(PARAM_PAGE) ||
				key.equals(PARAM_QUERY_TYPE) ||
				key.equals(PARAM_QUERY_PARENT) ||
				key.equals(PARAM_FETCH_CONCEPT_MODE)
				))
			this.defaultParams.put(key, value);
		else
			System.out.println("OnkiURLGenerator.setDefault(): No such key: "+key);
	}
	
	public String baseurl() {
		return this.baseurl;	
	}

	public String encoding() {
		return this.encoding;
	}
	
	// Methods for generating URLs ------------------------------------------------------
	
	/**
	 * Get the url of the front page (reset everything else than the language).
	 */
	
	public OnkiURL getFrontPageUrl() {
		Map<String,String> params = new HashMap<String,String>(this.defaultParams);
		params.put(PARAM_LANG, this.getParam(PARAM_LANG)); // keep the language, reset everything else
		return this.createOnkiURL(this.generateURLparams(params, this.defaultParams.keySet()));
	}
	
	/**
	 * Get the current url.
	 */
	
	public OnkiURL getCurrentUrl() {
		return this.createOnkiURL(this.generateURLparams(false));
	}
	
	/**
	 * Get the url of the default page, keep the other params.
	 */
	
	public OnkiURL getPageUrl() {
		return this.getPageUrl(null);
	}
	
	/**
	 * Get the url of the given page.
	 */
	
	public OnkiURL getPageUrl(String page) {
		return getPageUrl(page, null);
	}
	
	/**
	 * Get the url of the given page, keep the other params.
	 * @param page
	 * @param set of keys that should be filtered
	 */
	
	public OnkiURL getPageUrl(String page, Set<String> filterout) {
		Map<String,String> params = new HashMap<String,String>();
		if (page == null)
			page = this.defaultParams.get(PARAM_PAGE);
		if (page != null)
			params.put(PARAM_PAGE, page);
		return this.createOnkiURL(this.generateURLparams(params, null, filterout));
	}
	
	/**
	 * Get the url of the default page without lang, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutLang() {
		return this.getPageUrlWithoutLang(null);
	}
	
	/**
	 * Get the url of the given page without lang, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutLang(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_LANG);
		return this.getPageUrl(page, filterout);
	}
	
	/**
	 * Get the url of the default page without the query string, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQuery() {
		return this.getPageUrlWithoutQuery(null);
	}

	/**
	 * Get the url of the given page without the query string, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQuery(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_QUERY);
		return this.getPageUrl(page, filterout);
	}
	
	/**
	 * Get the url of the default page without the query type, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryType() {
		return this.getPageUrlWithoutQueryType(null);
	}

	/**
	 * Get the url of the given page without the query type, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryType(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_QUERY_TYPE);
		return this.getPageUrl(page, filterout);
	}

	/**
	 * Get the url of the default page without the query parent concept, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryParent() {
		return this.getPageUrlWithoutQueryParent(null);
	}

	/**
	 * Get the url of the given page without the query parent concept, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryParent(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_QUERY_PARENT);
		return this.getPageUrl(page, filterout);
	}

	/**
	 * Get the url of the default page without the query group concept, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryGroup() {
		return this.getPageUrlWithoutQueryGroup(null);
	}
	
	/**
	 * Get the url of the given page without the query group concept, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutQueryGroup(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_QUERY_GROUP);
		return this.getPageUrl(page, filterout);
	}
	
	/**
	 * Get the url of the default page without the group query string, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutGroupQuery() {
		return this.getPageUrlWithoutGroupQuery(null);
	}
	
	/**
	 * Get the url of the given page without the group query string, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithoutGroupQuery(String page) {
		Set<String> filterout = new HashSet<String>();
		filterout.add(PARAM_GROUP_QUERY);
		return this.getPageUrl(page, filterout);
	}	
	
	/**
	 * Get the url of the default page with the given concept, keep the other params.
	 */
	
	public OnkiURL getConceptPageUrl(String conceptUri) {
		return this.getPageUrlWithConcept(null, conceptUri);
	}
	
	/**
	 * Get the url of the given page with the given concept, keep the other params.
	 */

	public OnkiURL getPageUrlWithConcept(String page, String conceptUri) {
		return this.getPageUrlWithConcept(page, conceptUri, false);
	}

	/**
	 * Get the url of the given page with the given concept.
	 */
	
	public OnkiURL getPageUrlWithConcept(String page, String conceptUri, boolean clean) {
		Map<String,String> params = new HashMap<String,String>();
		if (page == null)
			page = this.defaultParams.get(PARAM_PAGE);
		if (page != null)
			params.put(PARAM_PAGE, page);
		params.put(PARAM_CONCEPT, conceptUri);
		
		if (clean)
			return this.createOnkiURL(this.generateURLparams(params, params.keySet()));
		else
			return this.createOnkiURL(this.generateURLparams(params));
	}
	
	/**
	 * Get the url of the default page with the given query group, keep the other params.
	 */
	
	public OnkiURL getQueryGroupPageUrl(String query) {
		return this.getPageUrlWithQueryGroup(null, query);
	}
	
	/**
	 * Get the url of the given page with the given query group, keep the other params.
	 */
	
	public OnkiURL getPageUrlWithQueryGroup(String page, String query) {
		Map<String,String> params = new HashMap<String,String>();
		if (page == null)
			page = this.defaultParams.get(PARAM_PAGE);
		if (page != null)
			params.put(PARAM_PAGE, page);
		params.put(PARAM_QUERY_GROUP, query);
		// when group is selected, all it's members are shown as search results
		params.put(PARAM_QUERY, "*");
		return this.createOnkiURL(this.generateURLparams(params));
	}
	
    /**
	 * Generate URL parameters.
	 */
	
	private Map<String,String[]> generateURLparams(boolean removeDefaults) {
		return this.generateURLparams(new HashMap<String,String>(), removeDefaults);
	}
	
	private Map<String,String[]> generateURLparams(Map<String,String> parameters) {
		return this.generateURLparams(parameters, true);
	}
	
	private Map<String,String[]> generateURLparams(Map<String,String> parameters, boolean removeDefaults) {
		return this.generateURLparams(parameters,null,removeDefaults);
	}

	private Map<String,String[]> generateURLparams(Map<String,String> params, Set<String> keys) {
		return this.generateURLparams(params,keys,true);
	}
	 
	private Map<String,String[]> generateURLparams(Map<String,String> parameters, Set<String> keys, boolean removeDefaults) {
		Map<String,String[]> urlparams = new HashMap<String,String[]>();
	
		// collect all potential keys to a set
		if (keys == null) {
			keys = this.getParameterKeys();
			if (parameters != null) 
				keys.addAll(parameters.keySet());
		}

		// iterate through the keys and generate the query
		for(String key : keys) {
			String[] value = null;

			if (key == null)
				continue;
			
			if (this.requestParams.containsKey(key)) {
				String[] stra = this.requestParams.get(key); // TODO: add support for multiple values
				if (stra != null && stra.length>0)
					//value = stra[0];
					value = stra;
			}

			// overwrite the value if we have a given parameter
			if (parameters != null && parameters.containsKey(key) && parameters.get(key) != null) // last condition added 3.7.2008
				{value = new String[1];
			    value[0] = parameters.get(key);}

			if (value == null && !removeDefaults && this.defaultParams != null &&
					this.defaultParams.containsKey(key)) {
				value = new String[1];
				value[0] = this.defaultParams.get(key);
			}

			if (value == null)
				continue;
			
			// don't add default values to 	the query
			if (removeDefaults && this.defaultParams != null && 
					this.defaultParams.containsKey(key) && 
					value.length == 1 && this.defaultParams.get(key).equals(value[0]))
				continue;

			urlparams.put(key, value);
		}

		return urlparams;
	}

	private Map<String,String[]> generateURLparams(Map<String,String> params, Set<String> keys, Set<String> filterout) {
		return this.generateURLparams(params,keys,filterout,true);
	}

	private Map<String,String[]> generateURLparams(Map<String,String> params, Set<String> keys, Set<String> filterout, boolean removeDefaults) {
		if (filterout != null) {
			Set<String> paramkeys = this.getParameterKeys();
			paramkeys.removeAll(filterout);
			if (keys != null)
				paramkeys.addAll(keys);
			paramkeys.addAll(params.keySet());
			return this.generateURLparams(params, paramkeys,removeDefaults);
		} else {
			return this.generateURLparams(params, keys,removeDefaults);
		}
	}
	
    /**
	 * Get all the parameter keys.
	 */
	
	private Set<String> getParameterKeys() {
		Set<String> keys = new HashSet<String>(this.requestParams.keySet());
		if (this.defaultParams != null)
			keys.addAll(this.defaultParams.keySet());
		return keys;
	}
	
    /**
	 * Get a specific parameter.
	 */
	
	private String getParam(String key) {
		if (this.requestParams.containsKey(key)) {
			String[] str = this.requestParams.get(key);
			if (str != null && str.length > 0)
				return str[0];
		}
		
		if (this.defaultParams.containsKey(key))
			return this.defaultParams.get(key);
		
		return null;
	}
	
    /**
	 * Create ONKIURL with parameters.
	 */
	
	private OnkiURL createOnkiURL(Map<String,String[]> params) {
		return new OnkiURL(params, this);
	}
	
	public String toString() {
		return "OnkiURLGenerator: currentURL: "+this.getCurrentUrl()+"\n" +
				"defaultparams: "+this.defaultParams;
	}
	
    /**
	 * Main method for testing.
	 */
	
	public static void main(String[] args) {
		Map<String,String[]> params = new HashMap<String,String[]>();
		params.put("q", new String[]{"elämä"});
		OnkiURLGenerator oug = new OnkiURLGenerator("http://joku.fi", params, "UTF-8");
		OnkiURL url = oug.getFrontPageUrl();
		System.out.println(url.toString());
	}
}