package fi.helsinki.cs.seco.onki.stats;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.vocabulary.OWL;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.vocabulary.RDFS;

/**
 * A Class for generating statistics of the used ontology.
 * 
 * @author vpkomula
 */
public class OnkiModelStatistics {

	private List nsStats;
	private static OnkiModelStatistics instance;
	private Logger logger;

	private OnkiModelStatistics(Model model) {
		logger = Logger.getLogger(OnkiModelStatistics.class);
		List stats = resolveStats(model);
		setNsStats(stats);
	}

	public static OnkiModelStatistics getInstance(Model model) {
		if (model != null && instance == null) {
			instance = new OnkiModelStatistics(model);
		}
		return instance;
	}

	public void setLogger() {
		logger = Logger.getLogger(OnkiModelStatistics.class);
	}	
	
    /**
	 * Getters and setters.
	 */
	
	public List getNsStats() {
		return nsStats;
	}

	public void setNsStats(List nsStats) {
		if (nsStats != null) {
			Collections.sort(nsStats);
		}
		this.nsStats = nsStats;
	}

	public int getTotalClassCount() {
		int result=0;
		for (Iterator iter = nsStats.iterator(); iter.hasNext();) {
			OnkiNsStat stat = (OnkiNsStat) iter.next();
			result+= stat.getClassCount();
		}
		return result;
	}

	public int getTotalIndividualCount() {
		int result=0;
		for (Iterator iter = nsStats.iterator(); iter.hasNext();) {
			OnkiNsStat stat = (OnkiNsStat) iter.next();
			result+= stat.getIndividualCount();
		}
		return result;
	}

	public int getTotalPropertyCount() {
		int result=0;
		for (Iterator iter = nsStats.iterator(); iter.hasNext();) {
			OnkiNsStat stat = (OnkiNsStat) iter.next();
			result+= stat.getPropertyCount();
		}
		return result;
	}

	public int getTotalStatementCount() {
		int result=0;
		for (Iterator iter = nsStats.iterator(); iter.hasNext();) {
			OnkiNsStat stat = (OnkiNsStat) iter.next();
			result+= stat.getStatementCount();
		}
		return result;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (Iterator iter = nsStats.iterator(); iter.hasNext();) {
			OnkiNsStat stat = (OnkiNsStat) iter.next();
			sb.append(stat).append(System.getProperty("line.separator"));
		}

		sb.append("Totals\t");
		sb.append("stmt:").append(getTotalStatementCount());
		sb.append(", Cls:").append(getTotalClassCount());
		sb.append(", Indiv:").append(getTotalIndividualCount());
		sb.append(", Prop:").append(getTotalPropertyCount());

		return sb.toString();
	}

	/** 
	 * Resolves the statistics of the ontology, e.g. amount of statements, classes, instances.
	 *
	 * @param model
	 * @return
	 */

	 private List resolveStats(Model model) {
		 List nsStatList = new ArrayList();
		 Hashtable<String,OnkiNsStat> statTable = new Hashtable<String,OnkiNsStat>();

		 //	Resolve used namespace
		 Map<String,String> prefixMap = model.getNsPrefixMap();

		 Set<String> keySet = prefixMap.keySet();
		 for (Iterator iter = keySet.iterator(); iter.hasNext(); ) {
			 String nsPrefix = (String) iter.next();
			 String ns = prefixMap.get(nsPrefix);
			 OnkiNsStat nsStat = new OnkiNsStat(ns, nsPrefix);
			 statTable.put(ns, nsStat);
		 }

		 //	Resolve Statements
		 for (StmtIterator iter = model.listStatements(); iter.hasNext();) {
			 Statement stmt = iter.nextStatement();
			 if (stmt.getSubject() != null && !stmt.getSubject().isAnon()) {
				 String ns = stmt.getSubject().getNameSpace();
				 if (statTable.containsKey(ns)) {
					 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
					 nsStat.setStatementCount(nsStat.getStatementCount() + 1);
				 }
				 // if there's no namespace prefix for namespace let's add an empty prefix to statTable
				 else {
					 OnkiNsStat nsStat = new OnkiNsStat(ns, "");
					 statTable.put(ns, nsStat);
					 nsStat.setStatementCount(nsStat.getStatementCount() + 1);
				 }
			 }
		 }

		 //	Resolve Classes
		 //	1. Resolve Class types (rdfs, owl)
		 List types = new ArrayList();
		 //RDFNode node = model.getRDFNode(Node.create("http://www.w3.org/2000/01/rdf-schema#Class"));
		 RDFNode node = model.getRDFNode(Node.createURI("http://www.w3.org/2000/01/rdf-schema#Class"));
		 for (ResIterator iter = model.listSubjectsWithProperty(RDF.type, node); iter.hasNext();) {
			 Resource res = iter.nextResource();
			 types.add(res);
		 }

		 //	TODO : Is this a class, I don't think so, this is the upmost level!!!
		//node = model.getRDFNode(Node.create("http://www.w3.org/2000/01/rdf-schema#Resource"));
		 node = model.getRDFNode(Node.createURI("http://www.w3.org/2000/01/rdf-schema#Resource"));
		 for (ResIterator iter = model.listSubjectsWithProperty(RDF.type, node); iter.hasNext();) {
			 Resource res = iter.nextResource();
			 types.add(res);
		 }

		 //node = model.getRDFNode(Node.create("http://www.w3.org/2002/07/owl#Class"));
		 node = model.getRDFNode(Node.createURI("http://www.w3.org/2002/07/owl#Class"));
		 for (ResIterator iter = model.listSubjectsWithProperty(RDF.type, node); iter.hasNext();) {
			 Resource res = iter.nextResource();
			 types.add(res);
		 }

		 //	2. List all statements, which type is in the list..

		 for (Iterator iter = types.iterator(); iter.hasNext();) {
			 Resource typeRes = (Resource) iter.next();

			 for (ResIterator iter2 = model.listSubjectsWithProperty(RDF.type,typeRes); iter2.hasNext();) {
				 Resource res = (Resource) iter2.next();
				 String ns = res.getNameSpace();
				 if (ns != null && statTable.containsKey(ns)) {
					 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
					 nsStat.setClassCount(nsStat.getClassCount() + 1);
				 }
			 }
		 }

		 //	Resolve Individuals
		 boolean hasIndividuals = false;
		 // for some reason this returns 0 individuals
		 for (ExtendedIterator iter = model.listSubjectsWithProperty(RDF.type); iter.hasNext();) {
			 hasIndividuals = true;
			 Resource res = (Resource) iter.next();
			 boolean isIndividual = true;
			 for (StmtIterator typeIter = res.listProperties(RDF.type); typeIter.hasNext(); ) { 
				 if (types.contains(typeIter.nextStatement().getObject()))
					 isIndividual = false;
			 }
			 if (!isIndividual || res.hasProperty(RDF.type, RDFS.Class) || res.hasProperty(RDF.type, OWL.Class))
				 continue;
			 String ns = res.getNameSpace();
			 if (ns != null && statTable.containsKey(ns)) {
				 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
				 nsStat.setIndividualCount(nsStat.getIndividualCount() + 1);
			 }
		 }
		 // treat SKOS concepts as individuals if previous for loop returned 0 individuals
		 if (!hasIndividuals) {
			 Resource conceptRes = model.getResource("http://www.w3.org/2004/02/skos/core#Concept");
			 for (ExtendedIterator iter = model.listSubjectsWithProperty(RDF.type, conceptRes); iter.hasNext();) { // Jouni, 28.10.2008
				 Resource res = (Resource) iter.next();
				 String ns = res.getNameSpace();
				 if (statTable.containsKey(ns)) {
					 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
					 nsStat.setIndividualCount(nsStat.getIndividualCount() + 1);
				 }
			 }
		 }

		 //	Resolve Properties
		 for (ExtendedIterator iter = model.listSubjectsWithProperty(RDF.type, RDF.Property); iter.hasNext();) {
			 Resource prop = (Resource) iter.next();
			 String ns = prop.getNameSpace();
			 if (statTable.containsKey(ns)) {
				 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
				 nsStat.setPropertyCount(nsStat.getPropertyCount() + 1);
			 }
		 }

		 // Resolve Properties
		 for (ExtendedIterator iter = model.listSubjectsWithProperty(RDF.type, OWL.ObjectProperty); iter.hasNext();) {
			 Resource prop = (Resource) iter.next();
			 String ns = prop.getNameSpace();
			 if (statTable.containsKey(ns)) {
				 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(ns);
				 nsStat.setPropertyCount(nsStat.getPropertyCount() + 1);
			 }
		 }

		 //	Change Hashtable to List
		 for (Iterator iter = statTable.keySet().iterator();iter.hasNext();) {
			 String key =  (String) iter.next();
			 OnkiNsStat nsStat = (OnkiNsStat) statTable.get(key);
			 nsStatList.add(nsStat);
		 }

		 return nsStatList;
	 }
}