package fi.seco.semweb.service.indexing.tree;

import java.sql.Types;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import fi.seco.semweb.service.indexing.tree.model.IInputNode;
import fi.seco.semweb.service.indexing.tree.model.IndexedNode;
import fi.seco.semweb.util.iterator.ANonExceptingResultSetMappingIterator;
import fi.seco.semweb.util.sql.NonExceptingConnection;
import fi.seco.semweb.util.sql.NonExceptingPreparedStatement;
import fi.seco.semweb.util.sql.NonExceptingResultSet;

public class TreeIndexService<T> {

	@SuppressWarnings("unused")
	private static Log log = LogFactory.getLog(TreeIndexService.class);

	protected String treeTable;

	protected NonExceptingPreparedStatement addToTree;

	private NonExceptingPreparedStatement getAncestorNodes;
	private NonExceptingPreparedStatement getSubNodes;
	private NonExceptingPreparedStatement getRoot;

	protected NonExceptingConnection sqlconn;

	protected ArrayList<IndexedNode<T>> nodeMap;

	private NonExceptingPreparedStatement getNodeByExternalId;

	public TreeIndexService(String tablePrefix, NonExceptingConnection sqlconn) {
		this.treeTable = tablePrefix+"_tree";
		this.sqlconn = sqlconn;
		if (!sqlconn.getMetaData().getTables(sqlconn.getCatalog(), null, (treeTable).toUpperCase(), null).next()) {
			sqlconn.createStatement().executeUpdate(
					"CREATE TABLE " + treeTable + " (nodeid INT, nodeid_end INT, parent INT, externalid VARCHAR(255), info OBJECT)");
			sqlconn.createStatement().executeUpdate(
					"CREATE INDEX " + treeTable + "_nodeid_nodeid_end_index ON " + treeTable + " (nodeid,nodeid_end)");
//			sqlconn.createStatement().executeUpdate("CREATE INDEX " + treeTable + "_parent_index ON " + treeTable + " (parent)");
			sqlconn.createStatement().executeUpdate("CREATE INDEX " + treeTable + "_externalid_index ON " + treeTable + " (externalid)");
		}
		addToTree = sqlconn.prepareStatement("INSERT INTO " + treeTable + " (nodeid, nodeid_end, parent, externalid, info) VALUES (?,?,?,?,?)");
		getSubNodes = sqlconn.prepareStatement("SELECT nodeid, nodeid_end, externalid, info FROM " + treeTable + " WHERE parent = ?");
		getRoot = sqlconn.prepareStatement("SELECT nodeid_end, externalid, info FROM " + treeTable + " WHERE nodeid=0");
		getNodeByExternalId = sqlconn.prepareStatement("SELECT nodeid FROM " + tablePrefix + "_tree WHERE externalid=?");
		getAncestorNodes = sqlconn.prepareStatement("SELECT nodeid FROM " + tablePrefix + "_tree WHERE nodeid<? AND nodeid_end > ?");
	}

	public void indexTree(IInputNode<T> tree) {
		clearTreeIndex();
		int lastId = 0;
		for (IInputNode<T> c : tree.getSubNodes())
			lastId = processNodeToIndex(c, 0, lastId + 1);
		addToTree.setInt(1, 0);
		addToTree.setInt(2, lastId);
		addToTree.setNull(3, Types.INTEGER);
		addToTree.setString(4, tree.getId());
		addToTree.setObject(5, tree.getInfo());
		addToTree.executeUpdate();
		nodeMap = null;
	}

	private int processNodeToIndex(IInputNode<T> c, int parentId, int myId) {
		int lastId = myId;
		for (IInputNode<T> c2 : c.getSubNodes())
			lastId = processNodeToIndex(c2, myId, lastId + 1);
		addToTree.setInt(1, myId);
		addToTree.setInt(2, lastId);
		addToTree.setInt(3, parentId);
		addToTree.setString(4, c.getId());
		addToTree.setObject(5, c.getInfo());
		addToTree.executeUpdate();
		return lastId;
	}

	public void clearTreeIndex() {
		sqlconn.createStatement().executeUpdate("DELETE FROM " + treeTable);
		nodeMap = null;
	}

	public String getTreeTable() {
		return treeTable;
	}

	@SuppressWarnings("unchecked")
	private void processNodeFromIndex(IndexedNode<T> parent) {
		getSubNodes.setInt(1, parent.getNodeId());
		NonExceptingResultSet rs = getSubNodes.executeQuery();
		while (rs.next()) {
			IndexedNode<T> c = new IndexedNode<T>(rs.getInt(1), rs.getInt(2), rs.getString(3),(T) rs.getObject(4));
			parent.addSubNode(c);
			nodeMap.add(c.getNodeId(), c);
			processNodeFromIndex(c);
		}
	}

	@SuppressWarnings("unchecked")
	private void getTreeFromIndex() {
		nodeMap = new ArrayList<IndexedNode<T>>();
		NonExceptingResultSet rs = getRoot.executeQuery();
		if (!rs.next()) log.error("No tree found in index");
		IndexedNode<T> ret = new IndexedNode<T>(0, rs.getInt(1), rs.getString(2), (T) rs.getObject(3));
		nodeMap.add(0, ret);
		processNodeFromIndex(ret);
	}

	public IndexedNode<T> getTree() {
		if (nodeMap == null) getTreeFromIndex();
		return nodeMap.get(0);
	}

	public List<IndexedNode<T>> getDescendantNodes(IndexedNode<T> c) {
		if (nodeMap == null) getTreeFromIndex();
		return nodeMap.subList(c.getNodeId(), c.getNodeid_end());
	}

	public Iterable<IndexedNode<T>> getAncestorNodes(IndexedNode<T> c) {
		if (nodeMap == null) getTreeFromIndex();
		getAncestorNodes.setInt(1, c.getNodeId());
		getAncestorNodes.setInt(2, c.getNodeid_end());
		return new ANonExceptingResultSetMappingIterator<IndexedNode<T>>(getAncestorNodes.executeQuery()) {
			@Override
			public IndexedNode<T> map() {
				return nodeMap.get(rs.getInt(1));
			}
		};

	}

	public ArrayList<IndexedNode<T>> getAllNodes() {
		if (nodeMap == null) getTreeFromIndex();
		return nodeMap;
	}

	public IndexedNode<T> getNodeByExternalId(String id) {
		if (nodeMap == null) getTreeFromIndex();
		getNodeByExternalId.setString(1, id);
		NonExceptingResultSet rs = getNodeByExternalId.executeQuery();
		return nodeMap.get(rs.getInt(1));
	}

	public Iterable<IndexedNode<T>> getNodesByExternalId(String id) {
		if (nodeMap == null) getTreeFromIndex();
		getNodeByExternalId.setString(1, id);
		return new ANonExceptingResultSetMappingIterator<IndexedNode<T>>(getNodeByExternalId.executeQuery()) {
			@Override
			public IndexedNode<T> map() {
				return nodeMap.get(rs.getInt(1));
			}
		};
	}

	public IndexedNode<T> getNodeById(int i) {
		if (nodeMap == null) getTreeFromIndex();
		return nodeMap.get(i);
	}

}
