package fi.seco.semweb.service.lucenegator;

import java.io.IOException;

import org.apache.lucene.index.TermDocs;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.Weight;

/**
 * A simplified version of TermScorer used by CategoryQuery. Most of the code is borrowed
 * from TermScorer.
 * @author oisuomin
 *
 */

public class CategoryScorer extends Scorer {

	private Weight weight;
	private TermDocs termDocs;
	private byte[] norms;
	private float weightValue;
	private int pointer;
	private int pointerMax;
	private int doc;
	private final int[] docs = new int[32]; // buffered doc numbers
	private final int[] freqs = new int[32]; // buffered term freqs

	public CategoryScorer(Weight weight, TermDocs td, Similarity similarity,
			byte[] norms) {
		super(similarity);
		this.weight = weight;
		this.termDocs = td;
		this.norms = norms;
		this.weightValue = weight.getValue();
	}

	@Override
	public boolean next() throws IOException {
		pointer++;
		if (pointer >= pointerMax) {
			pointerMax = termDocs.read(docs, freqs); // refill buffer
			if (pointerMax != 0)
				pointer = 0;
			else {
				termDocs.close(); // close stream
				doc = Integer.MAX_VALUE; // set to sentinel value
				return false;
			}
		}
		doc = docs[pointer];
		return true;
	}

	@Override
	public int doc() { return doc; }

	@Override
	public float score() throws IOException {
	    int f = freqs[pointer];
	    float raw = f * weightValue;
	    return raw * Similarity.decodeNorm(norms[doc]); // normalize for field
	}

	@Override
	public boolean skipTo(int target) throws IOException {
	    // first scan in cache
	    for (pointer++; pointer < pointerMax; pointer++)
			if (docs[pointer] >= target) {
			    doc = docs[pointer];
			    return true;
			  }

	    // not found in cache, seek underlying stream
	    boolean result = termDocs.skipTo(target);
	    if (result) {
	      pointerMax = 1;
	      pointer = 0;
	      docs[pointer] = doc = termDocs.doc();
	      freqs[pointer] = termDocs.freq();
	    } else doc = Integer.MAX_VALUE;
	    return result;
	}

	@Override
	public Explanation explain(int doc) throws IOException {
	    CategoryQuery query = (CategoryQuery)weight.getQuery();
	    Explanation tfExplanation = new Explanation();
	    int tf = 0;
	    while (pointer < pointerMax) {
	      if (docs[pointer] == doc)
	        tf = freqs[pointer];
	      pointer++;
	    }
	    if (tf == 0) while (termDocs.next())
			if (termDocs.doc() == doc) tf = termDocs.freq();
	    termDocs.close();
	    tfExplanation.setValue(tf);
	    tfExplanation.setDescription("termFreq("+query.getTerm()+")="+tf);

	    return tfExplanation;
	}

	@Override
	public String toString() {
		return "scorer(" + weight + ")";
	}

}
