package fi.seco.semweb.service.lucenegator;

import java.util.List;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.search.DefaultSimilarity;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.highlight.WeightedTerm;
import org.apache.lucene.util.SmallFloat;

/**
 * A class that encodes category values for Documents. Each category value is stored by
 * encoding it as a field value up to N times, where N is the pre-defined accuracy of the
 * field. At the same time, the field is given a boost ratio of 1/N. This, together with
 * using CategorySimilarity, ensures that the relevance for a CategoryQuery for the field
 * (at least approximately) corresponds to the given weight.
 * @author oisuomin
 *
 */

public class CategoryWriter {
	private static final int DEFAULT_ACCURACY = 8;
	private String fieldName;
	private float boost;
	private Similarity similarity;
	
	/**
	 * Create a new CategoryWriter for the given field name (usually something like "catid")
	 * using the given accuracy. The accuracy is a positive integer that defines the
	 * maximum number of times a term will be repeated. Higher values slow down the indexing
	 * and increase the index size (due to term positions being stored); however, they should
	 * not influence query times. Powers of 2 are good values as they can be exactly represented
	 * in the 8-bit float encoding used in the index.
	 * @param fieldName the name of the category field
	 * @param accuracy the desired accuracy of the encoding
	 * @param similarity the Similarity implementation used to calculate lengthNorm
	 */
	
	public CategoryWriter(String fieldName, int accuracy, Similarity similarity) {
		if (fieldName == null)
			throw new IllegalArgumentException("fieldName cannot be null");
		if (accuracy <= 0)
			throw new IllegalArgumentException("accuracy must be a positive integer");
		this.fieldName = fieldName;
		this.similarity = similarity;
		// set the boost factor to a value that can be accurately encoded in the Lucene index
		this.boost = SmallFloat.byte315ToFloat(SmallFloat.floatToByte315(1.0f / accuracy));
	}
	
	/**
	 * Create a new CategoryWriter using DefaultSimilarity.
	 * @param fieldName the name of the category field
	 * @param accuracy the desired accuracy of the encoding
	 */
	
	public CategoryWriter(String fieldName, int accuracy) {
		this(fieldName, accuracy, new DefaultSimilarity());
	}
	
	/**
	 * Construct a new CategoryWriter for the given field name using a default accuracy
	 * parameter and DefaultSimilarity.
	 * @param fieldName the name of the category field
	 */
	
	public CategoryWriter(String fieldName) {
		this(fieldName, DEFAULT_ACCURACY);
	}
	
	/**
	 * Mark a Document as belonging to a category with the given weight. This method may
	 * call Document.add(Field) several times. If isFirst is true, the method sets the boost 
	 * value of the field so that the relevance will be kept between 0.0 and 1.0.
	 * @param doc the document to mark
	 * @param catid category id
	 * @param weight weight of the category
	 * @param isFirst is this the first time the method is called for this document
	 */
	
	private void write(Document doc, String catid, float weight, int numCategories, boolean isFirst) {
		int repeat = Math.round(weight / boost);
		for (int i = 0; i < repeat; ++i) {
			Field f = new Field(fieldName, catid, Field.Store.NO, Field.Index.UN_TOKENIZED);
			if (isFirst && i == 0) { // first time the method is called on this document
				f.setBoost(boost * similarity.lengthNorm(fieldName, numCategories));
			}
			doc.add(f);
		}
	}
	
	/**
	 * Mark a Document as belonging to the given categories using the supplied weights.
	 * The boost factor will be set using the given number of categories.
	 * This method must be called exactly once for a given document, otherwise the
	 * boost factors will be miscalculated.
	 * @param doc the document to mark
	 * @param categories categories (with id and weight) where the document belongs
	 * @param numCats number of categories to use for calculating boost
	 */

	public void setCategories(Document doc, List<WeightedTerm> categories, int numCats) {
		boolean isFirst = true;
		for (WeightedTerm cat : categories) {
			write(doc, cat.getTerm(), cat.getWeight(), numCats, isFirst);
			isFirst = false;
		}
	}

	/**
	 * Mark a Document as belonging to the given categories using the supplied weights.
	 * The boost factor will be calculated from the number of categories in the given list.
	 * This method must be called exactly once for a given document, otherwise the
	 * boost factors will be miscalculated.
	 * @param doc the document to mark
	 * @param categories categories (with id and weight) where the document belongs
	 */

	public void setCategories(Document doc, List<WeightedTerm> categories) {
		setCategories(doc, categories, categories.size());
	}
}
