package fi.seco.semweb.service.multimedia;

import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import fi.seco.semweb.service.multimedia.model.IImageScalingSpec;


public class ImageScaler {
	
	private Log log = LogFactory.getLog(ImageScaler.class);
	
	private String format;
	private int width;
	private int height;
	private boolean enlarge;
	private boolean retainAspect;
	private final static String ENLARGE_DEFAULT = "true";
	private final static String RETAIN_ASPECT_DEFAULT = "true";
	
	public ImageScaler(IImageScalingSpec spec) {
		// TODO this can be done with some JavaBean stuff. Find out...
		this.setWidth(Integer.valueOf(spec.getWidth()).intValue());
		this.setHeight(Integer.valueOf(spec.getHeight()).intValue());
		//FIXME get the real parameters!!
		this.setEnlarge(new Boolean(ImageScaler.ENLARGE_DEFAULT).booleanValue());
		this.setRetainAspect(new Boolean(ImageScaler.RETAIN_ASPECT_DEFAULT).booleanValue());
	}
	
	/** 
     * Returns the affine transform that implements the scaling.
     * The behavior is the following: if both the new width and height values
     * are positive, the image is rescaled according to these new values and
     * the original aspect ration is lost.
     * Otherwise, if one of the two parameters is zero or negative, the
     * aspect ratio is maintained and the positive parameter indicates the
     * scaling.
     * If both new values are zero or negative, no scaling takes place (a unit
     * transformation is applied).
     */
    private AffineTransform getTransform(double ow, double oh, double nw, double nh) {
        double wm = 1.0d;
        double hm = 1.0d;
        if (nw > 0 && nh > 0 && retainAspect) {
            if (ow>oh) {
                wm = nw / ow;
                hm = wm;
            } else {
                hm = nh / oh;
                wm = hm;
            }
        } else {
            if (nw > 0) {
                wm = nw / ow;
                if (nh > 0) {
                    hm = nh / oh;
                } else {
                    hm = wm;
                }
            } else {
                if (nh > 0) {
                    hm = nh / oh;
                    wm = hm;
                }
            }
        }

        if (!enlarge) {
            if (retainAspect) {
                if (wm>1.0 || hm>1.0) {
                    hm=1.0;
                    wm=1.0; 
                }
            } else {
                if (wm>1.0) {
                    wm=1.0;
                }
                if (hm>1.0) {
                    hm=1.0;
                }
            }
        }
        return new AffineTransform(wm, 0.0d, 0.0d, hm, 0.0d, 0.0d);
    }

    /**
     * Scales an image and returns the scaled image
     * 
     * @param originalImage
     * @return
     * @throws IOException
     */
	public BufferedImage scale(BufferedImage originalImage) {
    	//if (format!=null) {
    		BufferedImage scaledImage;
			if (width > 0 || height > 0) {
				log.info("image " + ((width==0)?"?":Integer.toString(width))
									  + "x" + ((height==0)?"?":Integer.toString(height)));
	            Raster original = originalImage.getRaster();
	            double ow = original.getWidth();
            	double oh = original.getHeight();
            	log.info("original width: " + ow);
            	log.info("original height: " + oh);
            	AffineTransformOp filter = new AffineTransformOp(getTransform(ow, oh, width, height), AffineTransformOp.TYPE_BILINEAR);
            	WritableRaster scaled = filter.createCompatibleDestRaster(original);
            	filter.filter(original, scaled);
            	log.info("Image filtered.");
            	scaledImage = new BufferedImage(originalImage.getColorModel(),scaled,false,null);
            	log.info("Image scaled.");
			} else {
				scaledImage = originalImage;
			}
			return scaledImage;
        /*} else {
            // only read the resource - no modifications requested
            log.debug("passing original resource");
            return originalImage;
        }*/
    }

	public boolean isEnlarge() {
		return enlarge;
	}

	public void setEnlarge(boolean enlarge) {
		this.enlarge = enlarge;
	}

	public String getFormat() {
		return format;
	}

	public void setFormat(String format) {
		this.format = format;
	}

	public int getHeight() {
		return height;
	}

	public void setHeight(int height) {
		this.height = height;
	}

	public boolean isRetainAspect() {
		return retainAspect;
	}

	public void setRetainAspect(boolean retainAspect) {
		this.retainAspect = retainAspect;
	}

	public int getWidth() {
		return width;
	}

	public void setWidth(int width) {
		this.width = width;
	}
}
