package fi.seco.semweb.service.resourcelinking;


import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.hp.hpl.jena.rdf.model.Resource;

import fi.seco.semweb.model.WeightedResourceObject;

/**
 * Links resources by chaining other resource linkers. The first resource linker is given the
 * initial resource; the second operates on the results of the first, and so on. The result is
 * the combined results of the last resource linkers in the chain, which may contain duplicates.
 * Weights of the result set resources are calculated my multiplying the weights returned by the 
 * resource linkers.
 * @author oisuomin
 *
 */

public class ChainedResourceLinker implements IResourceLinker {

	private List<IResourceLinker> linkers;
	
	/**
	 * Create a new ChainedResourceLinker, specifying a chain of subordinate resource linkers.
	 * @param linkers list of subordinate resource linkers
	 */

	public ChainedResourceLinker(List<IResourceLinker> linkers) {
		this.linkers = linkers;
	}
	
	/**
	 * Convenience method for creating a ChainedResourceLinker for two resource linkers.
	 * @param first first subordinate resource linker
	 * @param second second subordinate resource linker
	 */
	public ChainedResourceLinker(IResourceLinker first, IResourceLinker second) {
		linkers = new ArrayList<IResourceLinker>(2);
		linkers.add(first);
		linkers.add(second);
	}

	/**
	 * Convenience method for creating a ChainedResourceLinker for three resource linkers.
	 * @param first first subordinate resource linker
	 * @param second second subordinate resource linker
	 * @param third third subordinate resource linker
	 */
	public ChainedResourceLinker(IResourceLinker first, IResourceLinker second, IResourceLinker third) {
		linkers = new ArrayList<IResourceLinker>(3);
		linkers.add(first);
		linkers.add(second);
		linkers.add(third);
	}
	
	/**
	 * Convenience method for creating a ChainedResourceLinker for four resource linkers.
	 * @param first first subordinate resource linker
	 * @param second second subordinate resource linker
	 * @param third third subordinate resource linker
	 * @param fourth fourth subordinate resource linker
	 */
	public ChainedResourceLinker(IResourceLinker first, IResourceLinker second, IResourceLinker third, IResourceLinker fourth) {
		linkers = new ArrayList<IResourceLinker>(4);
		linkers.add(first);
		linkers.add(second);
		linkers.add(third);
		linkers.add(fourth);
	}
	
	private Iterable<WeightedResourceObject> traverse(Resource resource, int linkerIdx, float weight) {
		List<WeightedResourceObject> result = new LinkedList<WeightedResourceObject>();

		IResourceLinker linker = linkers.get(linkerIdx);
		// if we're at the last linker, return its result, modifying the weights based on
		// previous steps
		if (linkerIdx == linkers.size() - 1) {
			for (WeightedResourceObject res : linker.getLinkedResources(resource))
				result.add(new WeightedResourceObject(res.getResource(), weight * res.getWeight()));
			return result;
		}
			
		// otherwise, recurse to all branches and merge their results
		for (WeightedResourceObject step : linker.getLinkedResources(resource))
			for (WeightedResourceObject res : traverse(step.getResource(), linkerIdx + 1, weight * step.getWeight()))
				result.add(res);

		return result;
	}
	
	public Iterable<WeightedResourceObject> getLinkedResources(Resource resource) {
		return traverse(resource, 0, 1.0f);
	}

}
