package fi.seco.semweb.service.resourcelinking;

import java.util.Collection;
import java.util.HashSet;

import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;

import fi.seco.semweb.model.WeightedResourceObject;

/**
 * Links resources by recursively following an RDF property. The original resource itself is part of
 * the returned result with weight 1.0. For each additional step, the weight of returned resources is
 * multiplied by the weightFactor given as parameter. Note that the same resource may be returned
 * several times, possibly with different weights; for a method of eliminating duplicates, see
 * MaxWeightResourceLinker. 
 * @author oisuomin
 * @see MaxWeightResourceLinker
 *
 */

public class TransitiveResourceLinker implements IResourceLinker {

	private Property property;
	private float weightFactor;

	/**
	 * Create a new TransitiveResourceLinker
	 * @param prop the property to traverse
	 * @param weightFactor weight factor specifying the amount weights diminish by when traversing
	 */
	
	public TransitiveResourceLinker(Property prop, float weightFactor) {
		this.property = prop;
		this.weightFactor = weightFactor;
	}
	
	public Iterable<WeightedResourceObject> getLinkedResources(Resource resource) {
		Collection<WeightedResourceObject> broader = new HashSet<WeightedResourceObject>();
		
		// the resource itself is always included, with weight factor 1.0
		broader.add(new WeightedResourceObject(resource, 1.0f));
		
		StmtIterator iter = resource.listProperties(property);
		
		while (iter.hasNext()) {
			Statement parentStatement = iter.nextStatement();
			Resource parent = (Resource) parentStatement.getObject();
			
			// add broaders of the parent concept recursively, reducing their weight by weightFactor
			for (WeightedResourceObject pb : getLinkedResources(parent)) {
				broader.add(new WeightedResourceObject(pb.getResource(), pb.getWeight() * weightFactor));
			}
		}
		return broader;

	}

}
