package fi.seco.semweb.service.search;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.Weight;
import org.apache.lucene.util.ToStringUtils;

import fi.seco.semweb.service.indexing.item.ICacheEventListener;
import fi.seco.semweb.service.indexing.item.ItemIndexService;
import fi.seco.semweb.service.search.query.ILuceneSelectorState;
import fi.seco.semweb.service.search.query.INativeSelectorState;
import fi.seco.semweb.service.search.query.IQueryPart;
import fi.seco.semweb.service.search.query.ISQLSelectorState;
import fi.seco.semweb.service.search.query.IntersectionQuery;
import fi.seco.semweb.service.search.query.UnionQuery;
import fi.seco.semweb.service.search.resultproducer.IResultProducer;
import fi.seco.semweb.service.search.resultproducer.IntersectionResultProducer;
import fi.seco.semweb.service.search.resultproducer.LuceneQueryResultProducer;
import fi.seco.semweb.service.search.resultproducer.SQLQueryResultProducer;
import fi.seco.semweb.util.sql.NonExceptingConnection;
import fi.seco.semweb.util.sql.NonExceptingPreparedStatement;
import fi.seco.semweb.util.sql.NonExceptingResultSet;
import gnu.trove.TIntObjectIterator;

public class ItemSearchController implements ICacheEventListener {

	@SuppressWarnings("unused")
	private static final Log log = LogFactory.getLog(ItemSearchController.class);

	private ItemIndexService index;
	private NonExceptingConnection sqlconn;

	public ItemSearchController() {
		BooleanQuery.setMaxClauseCount(Integer.MAX_VALUE);
	}

	public void setIndexService(ItemIndexService index) {
		this.index=index;
		sqlconn=index.getSqlConn();
		NonExceptingResultSet rs = sqlconn.getMetaData().getTables(sqlconn.getCatalog(),null,"TEMP"+sqlconn.getMetaData().getSearchStringEscape()+"_%",null);
		while (rs.next()) removeTempTable(rs.getString(3));
		index.setListener(this);
	}

	public ItemIndexService getIndex() {
		return index;
	}

	public IResultProducer planQuery(IQueryPart s) {
		if (s instanceof ISQLSelectorState) return new SQLQueryResultProducer(this,Collections.singleton((ISQLSelectorState)s));
		else if (s instanceof ILuceneSelectorState) return new LuceneQueryResultProducer(index,Collections.singleton((ILuceneSelectorState) s));
		else if (s instanceof IntersectionQuery) return handleIntersection((IntersectionQuery)s);
		else if (s instanceof INativeSelectorState) return ((INativeSelectorState) s).getResultProducer();
		else log.error("Attempted to handle an unknown IQueryPart "+s);
		return null;
	}

	private IResultProducer handleIntersection(IntersectionQuery iq) {
		Collection<ISQLSelectorState> sqlSelectors = new ArrayList<ISQLSelectorState>();
		Collection<ILuceneSelectorState> luceneSelectors = new ArrayList<ILuceneSelectorState>();
		List<IResultProducer> nativeSelectors = new ArrayList<IResultProducer>();
		ResultSet rs = null;
		for (IQueryPart s : iq.getParts())
			if (s instanceof ResultSet) { if (rs==null) rs=(ResultSet)s; else log.error("Intersections currently support only one ResultSet part, see SQLSelector for reason why"); }
			else if (s instanceof ISQLSelectorState) sqlSelectors.add((ISQLSelectorState) s);
			else if (s instanceof ILuceneSelectorState) luceneSelectors.add((ILuceneSelectorState) s);
			else if (s instanceof UnionQuery) nativeSelectors.add(handleUnion((UnionQuery)s));
			else if (s instanceof INativeSelectorState) nativeSelectors.add(((INativeSelectorState) s).getResultProducer());
			else log.error("Attempted to handle an unknown IQueryPart "+s);
		IntersectionResultProducer is = new IntersectionResultProducer();
		if (!sqlSelectors.isEmpty()) {
			SQLQueryResultProducer srp = new SQLQueryResultProducer(this, sqlSelectors);
			if (rs!=null) srp.initializeToFilter(rs);
			is.addResultProducer(srp);
		} else if (rs!=null) is.addResultProducer(rs.getResultProducer());
		if (!luceneSelectors.isEmpty()) is.addResultProducer(new LuceneQueryResultProducer(index, luceneSelectors));
		is.addResultProducers(nativeSelectors);
		return is;
	}

	private IResultProducer handleUnion(UnionQuery query) {
		log.error("Union's are not handled yet.");
		return null;
	}

	public ResultSet getResultSet(IQueryPart query) {
		Object s = query.getCacheKey();
		ResultSet results;
		if (s==null) {
			index.readLock().lock();
			results = new ResultSet(index,planQuery(query),null);
			index.readLock().unlock();
		} else {
			results = resultSetCache.get(s);
			if (results==null) {
				index.readLock().lock();
				results = new ResultSet(index,planQuery(query),s);
				index.readLock().unlock();
				resultSetCache.put(s,results);
			}
		}
		return results;
	}

	// Caching of result sets
	private int numEntriesInResultsetCache=0;

	protected LinkedHashMap<Object,ResultSet> resultSetCache = new LinkedHashMap<Object,ResultSet>(100,0.75F,true) {

		private static final long serialVersionUID = 1L;
		private static final int MAX_SIZE = 200000;

		@Override
		public ResultSet put(Object key, ResultSet value) {
			log.debug("Caching result (size: "+value.size()+") for cache key: "+key);
			numEntriesInResultsetCache+=value.size();
			if (numEntriesInResultsetCache>MAX_SIZE) {
				Iterator<ResultSet> it = values().iterator();
				while (numEntriesInResultsetCache>MAX_SIZE) {
					ResultSet m = it.next();
					numEntriesInResultsetCache-=m.size();
					it.remove();
				}
			}
			log.debug("Cache total size now: "+numEntriesInResultsetCache);
			return super.put(key,value);
		}

	};

	@SuppressWarnings("unchecked")
	public static Map<Object,Object> ensureMap(TIntObjectIterator i) {
		Map<Object,Object> tmp2 = (Map<Object, Object>) i.value();
		if (tmp2==null) {
			tmp2=new HashMap<Object,Object>();
			i.setValue(tmp2);
		}
		return tmp2;
	}

	public void invalidateCaches() {
		for (String r : tempTableCache.values()) removeTempTable(r);
		tempTableCache.clear();
		resultSetCache.clear();
		numEntriesInResultsetCache=0;
	}

	public class ResultSetQuery extends Query {

		private static final long serialVersionUID = 1L;

		ResultSet rs;

		public ResultSetQuery(ResultSet rs) {
			this.rs=rs;
		}

		@Override
		protected Weight createWeight(final Searcher searcher) throws IOException {
			return new Weight() {
				private static final long serialVersionUID = 1L;

				private float queryWeight=1.0f;

				private float queryNorm=1.0f;

				private float value=1.0f;

				public Query getQuery() {
					return ResultSetQuery.this;
				}

				public float getValue() {
					return value;
				}

				public float sumOfSquaredWeights() throws IOException {
					queryWeight = getBoost();
					return queryWeight * queryWeight;
				}

				public void normalize(float norm) {
					queryNorm = norm;
					queryWeight *= queryNorm;
					value = queryWeight;
				}

				public Scorer scorer(IndexReader reader) throws IOException {
					return new Scorer(getSimilarity(searcher)) {
						private int[] docs;
						private int pointer=-1;

						{
							this.docs = rs.ids();
						}

						@Override
						public boolean next() throws IOException {
							return ++pointer<docs.length;
						}

						@Override
						public int doc() {
							return docs[pointer];
						}

						@SuppressWarnings("unchecked")
						@Override
						public float score() throws IOException {
							return 1.0f;
						}

						@Override
						public boolean skipTo(int target) throws IOException {
						   do if (++pointer==docs.length) return false; while (target > docs[pointer]);
						   return true;
						}

						@Override
						public Explanation explain(int doc) throws IOException {
						    Explanation tfExplanation = new Explanation();
						    return tfExplanation;
						}

						@Override
						public String toString() {
							return "scorer()";
						}

					};
				}

				public Explanation explain(IndexReader reader, int doc)
						throws IOException {
					return new Explanation();
				}

				@Override
				public String toString() {
					return "weight(" + ResultSetQuery.this + ")";
				}
			};
		}

		/** Returns true iff <code>o</code> is equal to this. */
		@Override
		public boolean equals(Object o) {
			if (!(o instanceof ResultSetQuery))
				return false;
			ResultSetQuery other = (ResultSetQuery) o;
			return (this.getBoost() == other.getBoost())
					&& this.getResultSet().equals(other.getResultSet());
		}

		/** Returns a hash code value for this object. */
		@Override
		public int hashCode() {
			return Float.floatToIntBits(getBoost()) ^ getResultSet().hashCode();
		}

		@Override
		public String toString(String field) {
			return "ResultSetQuery("+field+")"+ToStringUtils.boost(getBoost());
		}

		public ResultSet getResultSet() {
			return rs;
		}

	}

	public Query getResultAsLuceneQuery(ResultSet rs) {
		return new ResultSetQuery(rs);
	}

	private long tempTableCount = 0;

	protected LinkedHashMap<Object,String> tempTableCache = new LinkedHashMap<Object,String>(100,0.75F,true) {

		private static final long serialVersionUID = 1L;
		private static final int MAX_ENTRIES = 200;

		@Override
		protected boolean removeEldestEntry(Map.Entry<Object,String> eldest) {
			if (size()> MAX_ENTRIES) {
				removeTempTable(eldest.getValue());
				return true;
			}
			return false;
		}
	};

	public void removeTempTable(String tableName) {
		sqlconn.createStatement().executeUpdate("DROP TABLE "+tableName);
	}

	public String createTempTable(int[] values) {
		String tempId="temp_"+tempTableCount++;
		sqlconn.createStatement().executeUpdate("CREATE TABLE "+tempId+" (id INT)");
		sqlconn.createStatement().executeUpdate("CREATE INDEX "+tempId+"_id_index ON "+tempId+" (id)");
		NonExceptingPreparedStatement addToTempTable=sqlconn.prepareStatement("INSERT INTO "+tempId+" (id) VALUES (?)");
		for (int key : values) {
			addToTempTable.setInt(1,key);
			addToTempTable.addBatch();
		}
		addToTempTable.executeBatch();
		return tempId;
	}

	public String getResultAsSQLTempTable(ResultSet rs) {
		Object s = rs.getCacheKey();
		String r = tempTableCache.get(s);
		if (r!=null) return r;
		r = createTempTable(rs.ids());
		tempTableCache.put(s,r);
		return r;
	}

	public NonExceptingConnection getSqlConn() {
		return sqlconn;
	}

}
